
from __future__ import annotations

from abc import abstractmethod
from typing import Optional, FrozenSet, Sequence
from typing_extensions import Protocol

from ..parsing.specials import Flag, ObjectId

__all__ = ['MailboxInterface']


class MailboxInterface(Protocol):
    """Describes a mailbox as it exists in an IMAP backend, whether or not it
    is currently selected.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def mailbox_id(self) -> ObjectId:
        """The mailbox object ID.

        This value must have no relationship to the mailbox name, e.g. a
        renamed mailbox should have the same ID but a deleted/re-created
        mailbox of the same name must have a different ID.

        See Also:
            `RFC 8474 4. <https://tools.ietf.org/html/rfc8474#section-4>`_

        """
        ...

    @property
    @abstractmethod
    def readonly(self) -> bool:
        """Whether the mailbox is read-only or read-write."""
        ...

    @property
    @abstractmethod
    def permanent_flags(self) -> FrozenSet[Flag]:
        """The permanent flags allowed in the mailbox."""
        ...

    @property
    @abstractmethod
    def session_flags(self) -> FrozenSet[Flag]:
        """The session flags allowed in the mailbox."""
        ...

    @property
    @abstractmethod
    def flags(self) -> FrozenSet[Flag]:
        """Set of all permanent and session flags available on the mailbox."""
        ...

    @property
    @abstractmethod
    def exists(self) -> int:
        """Number of total messages in the mailbox."""
        ...

    @property
    @abstractmethod
    def recent(self) -> int:
        """Number of recent messages in the mailbox."""
        ...

    @property
    @abstractmethod
    def unseen(self) -> int:
        """Number of unseen messages in the mailbox."""
        ...

    @property
    @abstractmethod
    def first_unseen(self) -> Optional[int]:
        """The sequence number of the first unseen message."""
        ...

    @property
    @abstractmethod
    def next_uid(self) -> int:
        """The predicted next message UID."""
        ...

    @property
    @abstractmethod
    def uid_validity(self) -> int:
        """The mailbox UID validity value."""
        ...




from __future__ import annotations

from abc import abstractmethod
from typing import TypeVar, Optional, Tuple, Sequence, FrozenSet, \
    Iterable, AsyncIterable
from typing_extensions import Protocol

from pymap.concurrent import Event
from pymap.flags import FlagOp
from pymap.interfaces.message import MessageT_co, CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials import ObjectId, SequenceSet
from pymap.parsing.specials.flag import get_system_flags, Flag, Deleted, Recent
from pymap.selected import SelectedSet, SelectedMailbox

__all__ = ['MailboxDataT', 'MailboxDataT_co',
           'MailboxDataInterface', 'MailboxSetInterface']

#: Type variable with an upper bound of :class:`MailboxDataInterface`.
MailboxDataT = TypeVar('MailboxDataT', bound='MailboxDataInterface')

#: Covariant type variable with an upper bound of
#: :class:`MailboxDataInterface`.
MailboxDataT_co = TypeVar('MailboxDataT_co', bound='MailboxDataInterface',
                          covariant=True)


class MailboxDataInterface(Protocol[MessageT_co]):
    """Manages the messages and metadata associated with a single mailbox."""

    @property
    @abstractmethod
    def mailbox_id(self) -> ObjectId:
        """The mailbox object ID.

        See Also:
            :attr:`~pymap.interfaces.mailbox.MailboxInterface.mailbox_id`

        """
        ...

    @property
    @abstractmethod
    def readonly(self) -> bool:
        """Whether the mailbox is read-only or read-write."""
        ...

    @property
    @abstractmethod
    def uid_validity(self) -> int:
        """The mailbox UID validity value."""
        ...

    @property
    def permanent_flags(self) -> FrozenSet[Flag]:
        """The permanent flags allowed in the mailbox."""
        return get_system_flags() - {Recent}

    @property
    def session_flags(self) -> FrozenSet[Flag]:
        """The session flags allowed in the mailbox."""
        return frozenset({Recent})

    @property
    @abstractmethod
    def selected_set(self) -> SelectedSet:
        """The set of selected mailbox sessions currently active."""
        ...

    @abstractmethod
    async def update_selected(self, selected: SelectedMailbox, *,
                              wait_on: Event = None) -> SelectedMailbox:
        """Populates and returns the selected mailbox object with the state
        needed to discover updates.

        Args:
            selected: the selected mailbox object.
            wait_on: If given, block until this event signals or mailbox
                activity occurs.

        """
        ...

    @abstractmethod
    async def append(self, append_msg: AppendMessage, *,
                     recent: bool = False) -> MessageT_co:
        """Adds a new message to the end of the mailbox, returning a copy of
        message with its assigned UID.

        Args:
            append_msg: The new message data.
            recent: True if the message should be marked recent.

        """
        ...

    @abstractmethod
    async def copy(self: MailboxDataT, uid: int, destination: MailboxDataT, *,
                   recent: bool = False) -> Optional[int]:
        """Copies a message, if it exists, from this mailbox to the
        *destination* mailbox.

        Args:
            uid: The UID of the message to copy.
            destination: The destination mailbox.
            recent: True if the message should be marked recent.

        """
        ...

    @abstractmethod
    async def move(self: MailboxDataT, uid: int, destination: MailboxDataT, *,
                   recent: bool = False) -> Optional[int]:
        """Moves a message, if it exists, from this mailbox to the
        *destination* mailbox.

        Args:
            uid: The UID of the message to move.
            destination: The destination mailbox.
            recent: True if the message should be marked recent.

        """
        ...

    @abstractmethod
    async def get(self, uid: int, cached_msg: CachedMessage) -> MessageT_co:
        """Return the message with the given UID.

        Args:
            uid: The message UID.
            cached_msg: The last known cached message.

        """
        ...

    @abstractmethod
    async def update(self, uid: int, cached_msg: CachedMessage,
                     flag_set: FrozenSet[Flag], mode: FlagOp) -> MessageT_co:
        """Update the permanent flags of the message.

        Args:
            uid: The message UID.
            cached_msg: The last known cached message.
            flag_set: The set of flags for the update operation.
            flag_op: The mode to change the flags.

        """
        ...

    @abstractmethod
    async def delete(self, uids: Iterable[int]) -> None:
        """Delete messages with the given UIDs.

        Args:
            uids: The message UIDs.

        """
        ...

    @abstractmethod
    async def claim_recent(self, selected: SelectedMailbox) -> None:
        """Messages that are newly added to the mailbox are assigned the
        ``\\Recent`` flag in the current selected mailbox session.

        Args:
            selected: The selected mailbox session.

        """
        ...

    @abstractmethod
    async def cleanup(self) -> None:
        """Perform any necessary "housekeeping" steps. This may be a slow
        operation, and may run things like garbage collection on the backend.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.check_mailbox`

        """
        ...

    @abstractmethod
    async def snapshot(self) -> MailboxSnapshot:
        """Returns a snapshot of the current state of the mailbox."""
        ...

    async def find(self, seq_set: SequenceSet, selected: SelectedMailbox) \
            -> AsyncIterable[Tuple[int, MessageT_co]]:
        """Find the active message UID and message pairs in the mailbox that
        are contained in the given sequences set. Message sequence numbers
        are resolved by the selected mailbox session.

        Args:
            seq_set: The sequence set of the desired messages.
            selected: The selected mailbox session.

        """
        for seq, cached_msg in selected.messages.get_all(seq_set):
            msg = await self.get(cached_msg.uid, cached_msg)
            if msg is not None:
                yield (seq, msg)

    async def find_deleted(self, seq_set: SequenceSet,
                           selected: SelectedMailbox) -> Sequence[int]:
        """Return all the active message UIDs that have the ``\\Deleted`` flag.

        Args:
            seq_set: The sequence set of the possible messages.
            selected: The selected mailbox session.

        """
        session_flags = selected.session_flags
        return [msg.uid async for _, msg in self.find(seq_set, selected)
                if Deleted in msg.get_flags(session_flags)]


class MailboxSetInterface(Protocol[MailboxDataT_co]):
    """Manages the set of mailboxes available to the authenticated user."""

    @property
    @abstractmethod
    def delimiter(self) -> str:
        """The delimiter used in mailbox names to indicate hierarchy."""
        ...

    @abstractmethod
    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        """Add or remove the subscribed status of a mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.subscribe`
            :meth:`~pymap.interfaces.session.SessionInterface.unsubscribe`

        Args:
            name: The name of the mailbox.
            subscribed: True if the mailbox should be subscribed.

        """
        ...

    @abstractmethod
    async def list_subscribed(self) -> ListTree:
        """Return a list of all subscribed mailboxes.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.list_mailboxes`

        """
        ...

    @abstractmethod
    async def list_mailboxes(self) -> ListTree:
        """Return a list of all mailboxes.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.list_mailboxes`

        """
        ...

    @abstractmethod
    async def get_mailbox(self, name: str) -> MailboxDataT_co:
        """Return an existing mailbox by name.

        Args:
            name: The name of the mailbox.

        Raises:
            KeyError: The mailbox did not exist.

        """
        ...

    @abstractmethod
    async def add_mailbox(self, name: str) -> ObjectId:
        """Create a new mailbox, returning its object ID.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.create_mailbox`

        Args:
            name: The name of the mailbox.

        Raises:
            ValueError: The mailbox already exists.

        """
        ...

    @abstractmethod
    async def delete_mailbox(self, name: str) -> None:
        """Delete an existing mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.delete_mailbox`

        Args:
            name: The name of the mailbox.

        Raises:
            KeyError: The mailbox did not exist.

        """
        ...

    @abstractmethod
    async def rename_mailbox(self, before: str, after: str) -> None:
        """Rename an existing mailbox.

        See Also:
            :meth:`~pymap.interfaces.session.SessionInterface.rename_mailbox`

        Args:
            before: The name of the existing mailbox.
            after: The name of the destination mailbox.

        Raises:
            KeyError: The *before* mailbox does not exist.
            ValueError: The *after* mailbox already exists.

        """
        ...




from __future__ import annotations

from typing import Optional, Tuple, Sequence, Dict

from pymap.filter import PluginFilterSet

__all__ = ['FilterSet']


class FilterSet(PluginFilterSet[bytes]):

    def __init__(self) -> None:
        super().__init__('sieve', bytes)
        self._filters: Dict[str, bytes] = {}
        self._active: Optional[str] = None

    async def put(self, name: str, value: bytes) -> None:
        self._filters[name] = value

    async def delete(self, name: str) -> None:
        if name not in self._filters:
            raise KeyError(name)
        elif name == self._active:
            raise ValueError(name)
        del self._filters[name]

    async def rename(self, before_name: str, after_name: str) -> None:
        if before_name not in self._filters:
            raise KeyError(before_name)
        elif after_name in self._filters:
            raise KeyError(after_name)
        self._filters[after_name] = self._filters[before_name]
        del self._filters[before_name]
        if self._active == before_name:
            self._active = after_name

    async def clear_active(self) -> None:
        self._active = None

    async def set_active(self, name: str) -> None:
        if name not in self._filters:
            raise KeyError(name)
        else:
            self._active = name

    async def get(self, name: str) -> bytes:
        return self._filters[name]

    async def get_active(self) -> Optional[bytes]:
        if self._active is None:
            return None
        else:
            return self._filters[self._active]

    async def get_all(self) -> Tuple[Optional[str], Sequence[str]]:
        return self._active, list(self._filters.keys())




from __future__ import annotations

import binascii
import random as _random
from typing import Optional, Tuple, AnyStr

from .. import Params, Parseable
from ..exceptions import NotParseable
from ...bytes import rev

__all__ = ['ObjectId']


class ObjectId(Parseable[bytes]):
    """Represents an object ID, used to identify a mailbox, message, or thread.

    An *object_id* value of ``None`` is a special case indicating that this
    object ID is not defined.

    Args:
        object_id: The object ID bytestring, or ``None``.

    See Also:
        `RFC 8474 <https://tools.ietf.org/html/rfc8474>`_

    """

    _pattern = rev.compile(br'[a-zA-Z0-9_-]{1,255}')

    def __init__(self, object_id: bytes = None) -> None:
        super().__init__()
        if object_id == b'':
            raise ValueError(object_id)
        self.object_id = object_id

    @property
    def value(self) -> bytes:
        """The object ID value."""
        if self.object_id is None:
            raise ValueError(self.object_id)
        return self.object_id

    @property
    def parens(self) -> bytes:
        """The object ID value surrounded by parentheses."""
        return b'(%b)' % (self.value, )

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[ObjectId, memoryview]:
        start = cls._whitespace_length(buf)
        match = cls._pattern.match(buf, start)
        if not match:
            raise NotParseable(buf)
        return cls(match.group(0)), buf[match.end(0):]

    @classmethod
    def new(cls, prefix: bytes, digest: bytes) -> ObjectId:
        """Return a new object ID from a hash digest.

        Args:
            prefix: The prefix for the object ID.
            digest: The hash digest.

        """
        return cls(prefix + binascii.hexlify(digest))

    @classmethod
    def new_mailbox_id(cls, digest: bytes) -> ObjectId:
        """Return a new mailbox ID from a hash digest.

        Args:
            digest: The hash digest.

        """
        return cls.new(b'F', digest)

    @classmethod
    def new_email_id(cls, digest: bytes) -> ObjectId:
        """Return a new email ID from a hash digest.

        Args:
            digest: The hash digest.

        """
        return cls.new(b'M', digest)

    @classmethod
    def new_thread_id(cls, digest: bytes) -> ObjectId:
        """Return a new thread ID from a hash digest.

        Args:
            digest: The hash digest.

        """
        return cls.new(b'T', digest)

    @classmethod
    def random(cls, prefix: bytes) -> ObjectId:
        """Return a new randomized object ID.

        Args:
            prefix: The prefix for the object ID.

        """
        return cls(b'%b%032x' % (prefix, _random.getrandbits(128)))

    @classmethod
    def random_mailbox_id(cls) -> ObjectId:
        """Return a new randomized mailbox ID."""
        return cls.random(b'F')

    @classmethod
    def random_email_id(cls) -> ObjectId:
        """Return a new randomized email ID."""
        return cls.random(b'M')

    @classmethod
    def random_thread_id(cls) -> ObjectId:
        """Return a new randomized thread ID."""
        return cls.random(b'T')

    @classmethod
    def maybe(cls, value: Optional[AnyStr]) -> ObjectId:
        """Return an object ID representing the string or bytestring value. If
        the input is empty or ``None``, the object ID returned will have
        :attr:`.not_defined` be true.

        Args:
            value: The object ID string or bytestring.

        """
        if not value:
            return cls(None)
        elif isinstance(value, str):
            return cls(value.encode('ascii', 'ignore'))
        else:
            return cls(value)

    def __eq__(self, other) -> bool:
        if isinstance(other, ObjectId):
            return self.object_id == other.object_id
        elif isinstance(other, bytes):
            return self.object_id == other
        return super().__eq__(other)

    def __hash__(self) -> int:
        return hash(self.object_id)

    def __str__(self) -> str:
        try:
            return self.parens.decode('ascii')
        except ValueError:
            return 'NIL'

    def __repr__(self) -> str:
        return '<ObjectId {0!s}>'.format(self)

    def __bytes__(self) -> bytes:
        return self.value




from __future__ import annotations

import os
import os.path
from abc import abstractmethod, ABCMeta
from contextlib import asynccontextmanager
from tempfile import NamedTemporaryFile
from typing import TypeVar, Type, Generic, Optional, IO, AsyncContextManager, \
    AsyncIterator

from pymap.concurrent import FileLock

__all__ = ['NoChanges', 'FileReadable', 'FileWriteable']

_RT = TypeVar('_RT', bound='FileReadable')
_WT = TypeVar('_WT', bound='FileWriteable')


class NoChanges(Exception):
    pass


class FileReadable(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def get_file(cls) -> str:
        ...

    @classmethod
    def get_lock(cls) -> Optional[str]:
        return None

    @classmethod
    @abstractmethod
    def get_default(cls: Type[_RT], base_dir: str) -> _RT:
        ...

    @classmethod
    @abstractmethod
    def open(cls: Type[_RT], base_dir: str, fp: IO[str]) -> _RT:
        ...

    def read(self, fp: IO[str]) -> None:
        pass

    @classmethod
    @asynccontextmanager
    async def _noop_lock(cls) -> AsyncIterator[None]:
        yield

    @classmethod
    def read_lock(cls, base_dir: str) -> AsyncContextManager[None]:
        lock_file = cls.get_lock()
        if lock_file:
            path = os.path.join(base_dir, lock_file)
            return FileLock(path).read_lock()
        else:
            return cls._noop_lock()

    @classmethod
    def write_lock(cls, base_dir: str) -> AsyncContextManager[None]:
        lock_file = cls.get_lock()
        if lock_file:
            path = os.path.join(base_dir, lock_file)
            return FileLock(path).write_lock()
        else:
            return cls._noop_lock()

    @classmethod
    def file_exists(cls, base_dir: str) -> bool:
        path = os.path.join(base_dir, cls.get_file())
        return os.path.exists(path)

    @classmethod
    def file_open(cls: Type[_RT], base_dir: str) -> _RT:
        path = os.path.join(base_dir, cls.get_file())
        try:
            with open(path, 'r') as in_file:
                return cls.open(base_dir, in_file)
        except FileNotFoundError:
            return cls.get_default(base_dir)

    @classmethod
    def file_read(cls: Type[_RT], base_dir: str) -> _RT:
        path = os.path.join(base_dir, cls.get_file())
        try:
            with open(path, 'r') as in_file:
                ret = cls.open(base_dir, in_file)
                ret.read(in_file)
                return ret
        except FileNotFoundError:
            return cls.get_default(base_dir)

    @classmethod
    def with_open(cls: Type[_RT], base_dir: str) -> _FileReadWith[_RT]:
        return _FileReadWith(base_dir, cls, True)

    @classmethod
    def with_read(cls: Type[_RT], base_dir: str) -> _FileReadWith[_RT]:
        return _FileReadWith(base_dir, cls, False)


class FileWriteable(FileReadable, metaclass=ABCMeta):

    @abstractmethod
    def get_dir(self) -> str:
        ...

    @abstractmethod
    def write(self, fp: IO[str]) -> None:
        ...

    @classmethod
    def delete(cls, base_dir: str) -> None:
        path = os.path.join(base_dir, cls.get_file())
        try:
            os.unlink(path)
        except OSError:
            pass

    def file_write(self) -> None:
        filename = self.get_file()
        base_dir = self.get_dir()
        tmp_dir = os.path.join(base_dir, 'tmp')
        path = os.path.join(base_dir, filename)
        with NamedTemporaryFile('w', dir=tmp_dir, delete=False) as tmp:
            self.write(tmp)
        tmp_path = os.path.join(base_dir, tmp.name)
        os.rename(tmp_path, path)

    @classmethod
    def with_write(cls: Type[_WT], base_dir: str) -> _FileWriteWith[_WT]:
        return _FileWriteWith(base_dir, cls)


class _FileReadWith(Generic[_RT]):

    def __init__(self, base_dir: str, cls: Type[_RT], only_open: bool) -> None:
        super().__init__()
        self._base_dir = base_dir
        self._cls = cls
        self._obj: Optional[_RT] = None
        self._only_open = only_open

    async def __aenter__(self) -> _RT:
        base_dir = self._base_dir
        cls = self._cls
        func = cls.file_open if self._only_open else cls.file_read
        async with cls.read_lock(base_dir):
            self._obj = obj = func(base_dir)
        return obj

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> bool:
        return False


class _FileWriteWith(Generic[_WT]):

    def __init__(self, base_dir: str, cls: Type[_WT]) -> None:
        super().__init__()
        self._base_dir = base_dir
        self._cls = cls
        self._exists = False
        self._lock: Optional[AsyncContextManager[None]] = None
        self._obj: Optional[_WT] = None

    async def _acquire_lock(self) -> None:
        self._lock = self._cls.write_lock(self._base_dir)
        await self._lock.__aenter__()

    async def _release_lock(self) -> None:
        lock = self._lock
        if not lock:
            raise RuntimeError()  # Lock not acquired.
        await lock.__aexit__(None, None, None)

    async def __aenter__(self) -> _WT:
        base_dir = self._base_dir
        cls = self._cls
        await self._acquire_lock()
        self._exists = cls.file_exists(base_dir)
        self._obj = obj = cls.file_read(base_dir)
        return obj

    def _write_obj(self) -> None:
        obj = self._obj
        if not obj:
            raise RuntimeError()  # Context manager not entered.
        obj.file_write()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> bool:
        try:
            if exc_type:
                if issubclass(exc_type, NoChanges):
                    if not self._exists:
                        self._write_obj()
                    return True
                else:
                    return False
            self._write_obj()
            return False
        finally:
            await self._release_lock()




from __future__ import annotations

from typing import Any, Optional, Mapping, Dict
from typing_extensions import Final

from pysasl import AuthenticationCredentials, HashInterface

from .config import IMAPConfig
from .exceptions import InvalidAuth

__all__ = ['UserMetadata']


class UserMetadata:
    """Contains user metadata including the password or hash.

    Args:
        config: The configuration object.
        password: The password string or hash digest.
        params: Additional metadata parameters associated with the user.

    """

    def __init__(self, config: IMAPConfig, password: Optional[str], *,
                 params: Mapping[str, str] = None) -> None:
        super().__init__()
        self.config: Final = config
        self.password: Final = password
        self.params: Final = params or {}

    @classmethod
    def from_dict(cls, config: IMAPConfig, data: Mapping[str, Any]) \
            -> UserMetadata:
        """Build a new :class:`UserMetadata` from a dictionary containing
        the password information. Fields must either be :func:`str` or
        :func:`int`, bytestrings should be hex-encoded.

        See Also:
            :meth:`.to_dict`

        Args:
            config: The configuration object.
            data: The password data dictionary.

        """
        params = dict(data)
        password = params.pop('password', None)
        return cls(config, password, params=params)

    def to_dict(self) -> Mapping[str, Any]:
        """Returns the password comparison data in a JSON-serializable
        dictionary for persistence and transfer.

        See Also:
            :meth:`.from_dict`

        """
        data: Dict[str, Any] = dict(self.params)
        if self.password:
            data['password'] = self.password
        return data

    async def check_password(self, creds: AuthenticationCredentials) -> None:
        """Check the given credentials against the known password comparison
        data. If the known data used a hash, then the equivalent hash of the
        provided secret is compared.

        Args:
            creds: The provided authentication credentials.

        Raises:
            :class:`~pymap.exceptions.InvalidAuth`

        """
        if self.password is None:
            raise InvalidAuth()
        hash_context = self.config.hash_context
        cpu_subsystem = self.config.cpu_subsystem
        fut = self._check_secret(creds, self.password, hash_context)
        if not await cpu_subsystem.execute(fut):
            raise InvalidAuth()

    async def _check_secret(self, creds: AuthenticationCredentials,
                            password: str,
                            hash_context: HashInterface) -> bool:
        return creds.check_secret(password, hash=hash_context)




from __future__ import annotations

from typing_extensions import Final

from aioredis import Redis  # type: ignore

from . import ScriptBase
from ..keys import CleanupKeys, NamespaceKeys, ContentKeys, MailboxKeys

__all__ = ['CleanupScripts']


class CleanupScripts:

    def __init__(self) -> None:
        super().__init__()
        self.namespace: Final = CleanupNamespace()
        self.mailbox: Final = CleanupMailbox()
        self.content: Final = CleanupContent()


class CleanupNamespace(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('cleanup_namespace')

    async def __call__(self, redis: Redis,
                       cl_keys: CleanupKeys, ns_keys: NamespaceKeys, *,
                       ttl: int) -> None:
        keys = [cl_keys.mailboxes, ns_keys.mailboxes, *ns_keys.keys]
        return await self.eval(redis, keys, [
            ttl, ns_keys.root.named['namespace']])


class CleanupMailbox(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('cleanup_mailbox')

    async def __call__(self, redis: Redis,
                       cl_keys: CleanupKeys, mbx_keys: MailboxKeys, *,
                       ttl: int) -> None:
        keys = [cl_keys.contents, mbx_keys.uids, mbx_keys.content,
                *mbx_keys.keys]
        return await self.eval(redis, keys, [
            ttl, mbx_keys.root.named['namespace'],
            mbx_keys.root.named['mailbox_id']])


class CleanupContent(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('cleanup_content')

    async def __call__(self, redis: Redis,
                       ns_keys: NamespaceKeys, ct_keys: ContentKeys, *,
                       ttl: int) -> None:
        keys = [ns_keys.content_refs, ct_keys.data]
        return await self.eval(redis, keys, [
            ttl, ct_keys.root.named['email_id']])




from __future__ import annotations

from typing_extensions import Final

from aioredis import Redis  # type: ignore

from . import ScriptBase
from ..keys import FilterKeys

__all__ = ['FilterScripts']


class FilterScripts:

    def __init__(self) -> None:
        super().__init__()
        self.get: Final = FilterGet()
        self.set_active: Final = FilterSetActive()
        self.delete: Final = FilterDelete()
        self.rename: Final = FilterRename()


class FilterGet(ScriptBase[bytes]):

    def __init__(self) -> None:
        super().__init__('filter_get')

    async def __call__(self, redis: Redis, fl_keys: FilterKeys, *,
                       name: bytes) -> bytes:
        keys = [fl_keys.names, fl_keys.data]
        args = [name]
        return await self.eval(redis, keys, args)


class FilterSetActive(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('filter_set_active')

    async def __call__(self, redis: Redis, fl_keys: FilterKeys, *,
                       name: bytes, active_name: bytes) -> None:
        keys = [fl_keys.names]
        args = [name, active_name]
        return await self.eval(redis, keys, args)


class FilterDelete(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('filter_delete')

    async def __call__(self, redis: Redis, fl_keys: FilterKeys, *,
                       name: bytes, active_name: bytes) -> None:
        keys = [fl_keys.names, fl_keys.data]
        args = [name, active_name]
        return await self.eval(redis, keys, args)


class FilterRename(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('filter_rename')

    async def __call__(self, redis: Redis, fl_keys: FilterKeys, *,
                       before_name: bytes, after_name: bytes) -> None:
        keys = [fl_keys.names]
        args = [before_name, after_name]
        return await self.eval(redis, keys, args)




import unittest

from pymap.parsing.response.specials import FlagsResponse, ExistsResponse, \
    RecentResponse, ExpungeResponse, FetchResponse, SearchResponse, \
    ESearchResponse, ListResponse, LSubResponse
from pymap.parsing.specials import FetchAttribute, FetchValue


class TestFlagsResponse(unittest.TestCase):

    def test_bytes(self):
        resp = FlagsResponse([br'\One', br'\Two'])
        self.assertEqual(b'* FLAGS (\\One \\Two)\r\n', bytes(resp))


class TestExistsResponse(unittest.TestCase):

    def test_bytes(self):
        resp = ExistsResponse(78)
        self.assertEqual(b'* 78 EXISTS\r\n', bytes(resp))


class TestRecentResponse(unittest.TestCase):

    def test_bytes(self):
        resp = RecentResponse(4)
        self.assertEqual(b'* 4 RECENT\r\n', bytes(resp))


class TestExpungeResponse(unittest.TestCase):

    def test_bytes(self):
        resp = ExpungeResponse(41)
        self.assertEqual(b'* 41 EXPUNGE\r\n', bytes(resp))


class TestFetchResponse(unittest.TestCase):

    def test_bytes(self):
        resp = FetchResponse(56, [
            FetchValue.of(FetchAttribute(b'KEY1'), b'VAL1')])
        self.assertEqual(b'* 56 FETCH (KEY1 VAL1)\r\n', bytes(resp))


class TestSearchResponse(unittest.TestCase):

    def test_bytes(self):
        resp = SearchResponse([4, 8, 15, 16, 23, 42])
        self.assertEqual(b'* SEARCH 4 8 15 16 23 42\r\n', bytes(resp))


class TestESearchResponse(unittest.TestCase):

    def test_bytes(self):
        resp = ESearchResponse(b'tag', True, {b'one': b'2', b'three': b'4'})
        self.assertEqual(b'* (TAG "tag") UID ONE 2 THREE 4\r\n', bytes(resp))


class TestListResponse(unittest.TestCase):

    def test_bytes(self):
        resp1 = ListResponse('inbox', '.', [])
        self.assertEqual(b'* LIST () "." INBOX\r\n', bytes(resp1))
        resp2 = ListResponse('Other.Stuff', '.',
                             [b'Marked', b'Noinferior', b'Noselect'])
        self.assertEqual(b'* LIST (\\Marked \\Noinferior \\Noselect) '
                         b'"." Other.Stuff\r\n', bytes(resp2))


class TestLSubResponse(unittest.TestCase):

    def test_bytes(self):
        resp1 = LSubResponse('inbox', '.', [])
        self.assertEqual(b'* LSUB () "." INBOX\r\n', bytes(resp1))
        resp2 = LSubResponse('Other.Stuff', '.',
                             [b'Marked', b'Noinferior', b'Noselect'])
        self.assertEqual(b'* LSUB (\\Marked \\Noinferior \\Noselect) '
                         b'"." Other.Stuff\r\n', bytes(resp2))




from __future__ import annotations

from email.headerregistry import HeaderRegistry, BaseHeader, \
    UnstructuredHeader, DateHeader, AddressHeader, SingleAddressHeader, \
    ContentDispositionHeader, ContentTransferEncodingHeader, ContentTypeHeader
from email.policy import SMTP
from typing import cast, Any, Optional, ClassVar, Iterable, Iterator, \
    Sequence, Mapping, Dict

__all__ = ['ParsedHeaders']

_Headers = Mapping[bytes, Sequence[Sequence[bytes]]]


class ParsedHeaders(Mapping[bytes, Sequence[BaseHeader]]):
    """The mapping of message headers, parsed on-demand using a
    :class:`~email.headerregistry.HeaderRegistry`. Also provides typed
    properties for standard headers used in IMAP processing.

    """

    _registry: ClassVar[HeaderRegistry] = HeaderRegistry()

    __slots__ = ['_headers', '_parsed']

    def __init__(self, headers: _Headers) -> None:
        super().__init__()
        self._headers = headers
        self._parsed: Dict[bytes, Sequence[BaseHeader]] = {}

    def __getitem__(self, name: bytes) -> Sequence[BaseHeader]:
        name_lower = name.lower()
        parsed = self._parsed.get(name_lower)
        if parsed is not None:
            return parsed
        values = self._headers[name_lower]
        self._parsed[name_lower] = parsed = list(self._parse(values))
        return parsed

    def __len__(self) -> int:
        return len(self._headers)

    def __contains__(self, name: Any) -> bool:
        if not isinstance(name, bytes):
            raise TypeError(name)
        return name.lower() in self._headers

    def __iter__(self) -> Iterator[bytes]:
        return iter(self._headers.keys())

    @classmethod
    def _parse(cls, values: Sequence[Sequence[bytes]]) \
            -> Iterable[BaseHeader]:
        for value in values:
            lines = [line.decode('ascii', 'surrogateescape')
                     for line in value]
            # TODO: Once typeshed merges this fix:
            #   https://github.com/python/typeshed/pull/4365
            # assign to hdr_name, hdr_value = ... instead.
            hdr_tuple = SMTP.header_source_parse(lines)
            yield cls._registry(hdr_tuple[0], hdr_tuple[1])

    def __repr__(self) -> str:
        return repr(dict(self))

    @property
    def content_type(self) -> Optional[ContentTypeHeader]:
        """The ``Content-Type`` header."""
        try:
            return cast(ContentTypeHeader, self[b'content-type'][0])
        except (KeyError, IndexError):
            return None

    @property
    def date(self) -> Optional[DateHeader]:
        """The ``Date`` header."""
        try:
            return cast(DateHeader, self[b'date'][0])
        except (KeyError, IndexError):
            return None

    @property
    def subject(self) -> Optional[UnstructuredHeader]:
        """The ``Subject`` header."""
        try:
            return cast(UnstructuredHeader, self[b'subject'][0])
        except (KeyError, IndexError):
            return None

    @property
    def from_(self) -> Optional[Sequence[AddressHeader]]:
        """The ``From`` header."""
        try:
            return cast(Sequence[AddressHeader], self[b'from'])
        except KeyError:
            return None

    @property
    def sender(self) -> Optional[Sequence[SingleAddressHeader]]:
        """The ``Sender`` header."""
        try:
            return cast(Sequence[SingleAddressHeader], self[b'sender'])
        except KeyError:
            return None

    @property
    def reply_to(self) -> Optional[Sequence[AddressHeader]]:
        """The ``Reply-To`` header."""
        try:
            return cast(Sequence[AddressHeader], self[b'reply-to'])
        except KeyError:
            return None

    @property
    def to(self) -> Optional[Sequence[AddressHeader]]:
        """The ``To`` header."""
        try:
            return cast(Sequence[AddressHeader], self[b'to'])
        except KeyError:
            return None

    @property
    def cc(self) -> Optional[Sequence[AddressHeader]]:
        """The ``Cc`` header."""
        try:
            return cast(Sequence[AddressHeader], self[b'cc'])
        except KeyError:
            return None

    @property
    def bcc(self) -> Optional[Sequence[AddressHeader]]:
        """The ``Bcc`` header."""
        try:
            return cast(Sequence[AddressHeader], self[b'bcc'])
        except KeyError:
            return None

    @property
    def in_reply_to(self) -> Optional[UnstructuredHeader]:
        """The ``In-Reply-To`` header."""
        try:
            return cast(UnstructuredHeader, self[b'in-reply-to'][0])
        except (KeyError, IndexError):
            return None

    @property
    def references(self) -> Optional[UnstructuredHeader]:
        """The ``References`` header."""
        try:
            return cast(UnstructuredHeader, self[b'references'][0])
        except (KeyError, IndexError):
            return None

    @property
    def message_id(self) -> Optional[UnstructuredHeader]:
        """The ``Message-Id`` header."""
        try:
            return cast(UnstructuredHeader, self[b'message-id'][0])
        except (KeyError, IndexError):
            return None

    @property
    def content_disposition(self) -> Optional[ContentDispositionHeader]:
        """The ``Content-Disposition`` header."""
        try:
            return cast(ContentDispositionHeader,
                        self[b'content-disposition'][0])
        except (KeyError, IndexError):
            return None

    @property
    def content_language(self) -> Optional[UnstructuredHeader]:
        """The ``Content-Language`` header."""
        try:
            return cast(UnstructuredHeader, self[b'content-language'][0])
        except (KeyError, IndexError):
            return None

    @property
    def content_location(self) -> Optional[UnstructuredHeader]:
        """The ``Content-Location`` header."""
        try:
            return cast(UnstructuredHeader, self[b'content-location'][0])
        except (KeyError, IndexError):
            return None

    @property
    def content_id(self) -> Optional[UnstructuredHeader]:
        """The ``Content-Id`` header."""
        try:
            return cast(UnstructuredHeader, self[b'content-id'][0])
        except (KeyError, IndexError):
            return None

    @property
    def content_description(self) -> Optional[UnstructuredHeader]:
        """The ``Content-Description`` header."""
        try:
            return cast(UnstructuredHeader, self[b'content-description'][0])
        except (KeyError, IndexError):
            return None

    @property
    def content_transfer_encoding(self) \
            -> Optional[ContentTransferEncodingHeader]:
        """The ``Content-Transfer-Encoding`` header."""
        try:
            return cast(ContentTransferEncodingHeader,
                        self[b'content-transfer-encoding'][0])
        except (KeyError, IndexError):
            return None




from __future__ import annotations

from typing import Optional, Tuple, Sequence, List

from aioredis import Redis, ReplyError  # type: ignore

from .keys import NamespaceKeys, FilterKeys
from .scripts.filter import FilterScripts
from pymap.filter import PluginFilterSet
from pymap.parsing.specials import ObjectId

__all__ = ['FilterSet']

_scripts = FilterScripts()


class FilterSet(PluginFilterSet[bytes]):

    _active_name = b''

    def __init__(self, redis: Redis, ns_keys: NamespaceKeys) -> None:
        super().__init__('sieve', bytes)
        self._redis = redis
        self._keys = FilterKeys(ns_keys)

    async def put(self, name: str, value: bytes) -> None:
        keys = self._keys
        name_bytes = name.encode('utf-8')
        key = ObjectId.random(b'S').value
        pipe = self._redis.pipeline()
        pipe.hset(keys.data, key, value)
        pipe.hset(keys.names, name_bytes, key)
        await pipe.execute()

    async def delete(self, name: str) -> None:
        try:
            await _scripts.delete(self._redis, self._keys,
                                  name=name.encode('utf-8'),
                                  active_name=self._active_name)
        except ReplyError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(name) from exc
            elif 'filter is active' in str(exc):
                raise ValueError(name) from exc
            raise

    async def rename(self, before_name: str, after_name: str) -> None:
        try:
            await _scripts.rename(self._redis, self._keys,
                                  before_name=before_name.encode('utf-8'),
                                  after_name=after_name.encode('utf-8'))
        except ReplyError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(before_name) from exc
            elif 'filter already exists' in str(exc):
                raise KeyError(after_name) from exc
            raise

    async def clear_active(self) -> None:
        keys = self._keys
        await self._redis.hdel(keys.names, self._active_name)

    async def set_active(self, name: str) -> None:
        try:
            await _scripts.set_active(self._redis, self._keys,
                                      name=name.encode('utf-8'),
                                      active_name=self._active_name)
        except ReplyError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(name) from exc
            raise

    async def get(self, name: str) -> bytes:
        try:
            return await _scripts.get(self._redis, self._keys,
                                      name=name.encode('utf-8'))
        except ReplyError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(name) from exc
            raise

    async def get_active(self) -> Optional[bytes]:
        try:
            return await _scripts.get(self._redis, self._keys,
                                      name=self._active_name)
        except ReplyError as exc:
            if 'filter not found' in str(exc):
                return None
            raise

    async def get_all(self) -> Tuple[Optional[str], Sequence[str]]:
        keys = self._keys
        sieve_names = await self._redis.hgetall(keys.names)
        active_key: Optional[bytes] = sieve_names.get(self._active_name)
        active_name: Optional[str] = None
        names: List[str] = []
        for name, key in sieve_names.items():
            if name != self._active_name:
                name_str = name.decode('utf-8')
                names.append(name_str)
                if key == active_key:
                    active_name = name_str
        return active_name, names




from __future__ import annotations

from abc import abstractmethod
from asyncio import shield
from typing import Generic, Tuple, Optional, FrozenSet, Iterable, \
    Sequence, List

from pymap.concurrent import Event
from pymap.config import IMAPConfig
from pymap.exceptions import MailboxNotFound, MailboxConflict, MailboxReadOnly
from pymap.flags import FlagOp, SessionFlags, PermanentFlags
from pymap.interfaces.filter import FilterSetInterface
from pymap.interfaces.message import MessageT
from pymap.interfaces.session import SessionInterface
from pymap.interfaces.users import UsersInterface
from pymap.mailbox import MailboxSnapshot
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials import SequenceSet, SearchKey, ObjectId, \
    FetchRequirement
from pymap.parsing.specials.flag import Flag, Seen
from pymap.parsing.response.code import AppendUid, CopyUid
from pymap.search import SearchParams, SearchCriteriaSet
from pymap.selected import SelectedMailbox

from .mailbox import MailboxDataInterface, MailboxSetInterface

__all__ = ['BaseSession']


class BaseSession(SessionInterface, Generic[MessageT]):
    """Base implementation of
    :class:`~pymap.interfaces.session.SessionInterface` intended for use by
    most backends.

    Args:
        owner: The logged-in user name.

    """

    def __init__(self, owner: str) -> None:
        super().__init__()
        self._owner = owner

    @property
    def owner(self) -> str:
        return self._owner

    @property
    def users(self) -> Optional[UsersInterface]:
        return None  # no user management by default

    @property
    @abstractmethod
    def config(self) -> IMAPConfig:
        ...

    @property
    @abstractmethod
    def mailbox_set(self) \
            -> MailboxSetInterface[MailboxDataInterface[MessageT]]:
        ...

    @property
    def filter_set(self) -> Optional[FilterSetInterface]:
        return None

    def close(self) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def _load_updates(self, selected: Optional[SelectedMailbox],
                            mbx: Optional[MailboxDataInterface[MessageT]]) \
            -> Optional[SelectedMailbox]:
        if selected:
            if not mbx or selected.mailbox_id != mbx.mailbox_id:
                try:
                    mbx = await self._get_selected(selected)
                except MailboxNotFound:
                    selected.set_deleted()
                    return selected
            return await mbx.update_selected(selected)
        return selected

    @classmethod
    def _pick_selected(cls, selected: Optional[SelectedMailbox],
                       mbx: MailboxDataInterface[MessageT]) \
            -> Optional[SelectedMailbox]:
        if selected and selected.mailbox_id == mbx.mailbox_id:
            return selected
        return mbx.selected_set.any_selected

    async def _get_mailbox(self, name: str, *, try_create: bool = False) \
            -> MailboxDataInterface[MessageT]:
        try:
            return await self.mailbox_set.get_mailbox(name)
        except KeyError as exc:
            raise MailboxNotFound(name, try_create=try_create) from exc

    async def _get_selected(self, selected: SelectedMailbox) \
            -> MailboxDataInterface[MessageT]:
        return await self._get_mailbox(selected.lookup)

    async def list_mailboxes(self, ref_name: str, filter_: str,
                             subscribed: bool = False,
                             selected: SelectedMailbox = None) \
            -> Tuple[Iterable[Tuple[str, Optional[str], Sequence[bytes]]],
                     Optional[SelectedMailbox]]:
        delimiter = self.mailbox_set.delimiter
        if filter_:
            if subscribed:
                list_tree = await self.mailbox_set.list_subscribed()
            else:
                list_tree = await self.mailbox_set.list_mailboxes()
            ret = [(entry.name, delimiter, entry.attributes)
                   for entry in list_tree.list_matching(ref_name, filter_)]
        else:
            ret = [("", delimiter, [b'Noselect'])]
        return ret, await self._load_updates(selected, None)

    async def get_mailbox(self, name: str, selected: SelectedMailbox = None) \
            -> Tuple[MailboxSnapshot, Optional[SelectedMailbox]]:
        try:
            mbx = await self.mailbox_set.get_mailbox(name)
        except KeyError as exc:
            raise MailboxNotFound(name) from exc
        snapshot = await mbx.snapshot()
        return snapshot, await self._load_updates(selected, mbx)

    async def create_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Tuple[ObjectId, Optional[SelectedMailbox]]:
        try:
            mailbox_id = await self.mailbox_set.add_mailbox(name)
        except ValueError as exc:
            raise MailboxConflict(name) from exc
        return mailbox_id, await self._load_updates(selected, None)

    async def delete_mailbox(self, name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        try:
            await self.mailbox_set.delete_mailbox(name)
        except KeyError as exc:
            raise MailboxNotFound(name) from exc
        return await self._load_updates(selected, None)

    async def rename_mailbox(self, before_name: str, after_name: str,
                             selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        try:
            await self.mailbox_set.rename_mailbox(before_name, after_name)
        except KeyError as exc:
            raise MailboxNotFound(before_name) from exc
        except ValueError as exc:
            raise MailboxConflict(after_name) from exc
        return await self._load_updates(selected, None)

    async def subscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        await self.mailbox_set.set_subscribed(name, True)
        return await self._load_updates(selected, None)

    async def unsubscribe(self, name: str, selected: SelectedMailbox = None) \
            -> Optional[SelectedMailbox]:
        await self.mailbox_set.set_subscribed(name, False)
        return await self._load_updates(selected, None)

    async def append_messages(self, name: str,
                              messages: Sequence[AppendMessage],
                              selected: SelectedMailbox = None) \
            -> Tuple[AppendUid, Optional[SelectedMailbox]]:
        mbx = await self._get_mailbox(name, try_create=True)
        if mbx.readonly:
            raise MailboxReadOnly(name)
        dest_selected = self._pick_selected(selected, mbx)
        uids: List[int] = []
        for append_msg in messages:
            msg = await mbx.append(append_msg, recent=not dest_selected)
            if dest_selected:
                dest_selected.session_flags.add_recent(msg.uid)
            uids.append(msg.uid)
        return (AppendUid(mbx.uid_validity, uids),
                await self._load_updates(selected, mbx))

    async def select_mailbox(self, name: str, readonly: bool = False) \
            -> Tuple[MailboxSnapshot, SelectedMailbox]:
        mbx = await self._get_mailbox(name)
        selected = SelectedMailbox(mbx.mailbox_id, readonly or mbx.readonly,
                                   PermanentFlags(mbx.permanent_flags),
                                   SessionFlags(mbx.session_flags),
                                   selected_set=mbx.selected_set,
                                   lookup=name)
        if not selected.readonly:
            await mbx.claim_recent(selected)
        snapshot = await mbx.snapshot()
        return snapshot, await mbx.update_selected(selected)

    async def check_mailbox(self, selected: SelectedMailbox, *,
                            wait_on: Event = None,
                            housekeeping: bool = False) -> SelectedMailbox:
        mbx = await self._get_selected(selected)
        if housekeeping:
            await shield(mbx.cleanup())
        return await mbx.update_selected(selected, wait_on=wait_on)

    async def fetch_messages(self, selected: SelectedMailbox,
                             sequence_set: SequenceSet, set_seen: bool) \
            -> Tuple[Iterable[Tuple[int, MessageT]], SelectedMailbox]:
        mbx = await self._get_selected(selected)
        ret: List[Tuple[int, MessageT]] = []
        for seq, cached_msg in selected.messages.get_all(sequence_set):
            if set_seen:
                msg = await mbx.update(cached_msg.uid, cached_msg,
                                       frozenset({Seen}), FlagOp.ADD)
            else:
                msg = await mbx.get(cached_msg.uid, cached_msg)
            if msg is not None:
                ret.append((seq, msg))
        return ret, await mbx.update_selected(selected)

    async def search_mailbox(self, selected: SelectedMailbox,
                             keys: FrozenSet[SearchKey]) \
            -> Tuple[Iterable[Tuple[int, MessageT]], SelectedMailbox]:
        mbx = await self._get_selected(selected)
        req = FetchRequirement.reduce(key.requirement for key in keys)
        ret: List[Tuple[int, MessageT]] = []
        params = SearchParams(selected,
                              disabled=self.config.disable_search_keys)
        search = SearchCriteriaSet(keys, params)
        async for seq, msg in mbx.find(search.sequence_set, selected):
            msg_content = await msg.load_content(req)
            if search.matches(seq, msg, msg_content):
                ret.append((seq, msg))
        return ret, await mbx.update_selected(selected)

    async def expunge_mailbox(self, selected: SelectedMailbox,
                              uid_set: SequenceSet = None) -> SelectedMailbox:
        if selected.readonly:
            raise MailboxReadOnly()
        mbx = await self._get_selected(selected)
        if uid_set is None:
            uid_set = SequenceSet.all(uid=True)
        expunge_uids = await mbx.find_deleted(uid_set, selected)
        await mbx.delete(expunge_uids)
        return await mbx.update_selected(selected)

    async def copy_messages(self, selected: SelectedMailbox,
                            sequence_set: SequenceSet,
                            mailbox: str) \
            -> Tuple[Optional[CopyUid], SelectedMailbox]:
        mbx = await self._get_selected(selected)
        dest = await self._get_mailbox(mailbox, try_create=True)
        if dest.readonly:
            raise MailboxReadOnly(mailbox)
        dest_selected = self._pick_selected(selected, dest)
        uids: List[Tuple[int, int]] = []
        for _, source_uid in selected.messages.get_uids(sequence_set):
            dest_uid = await mbx.copy(source_uid, dest,
                                      recent=not dest_selected)
            if dest_uid is not None:
                if dest_selected:
                    dest_selected.session_flags.add_recent(dest_uid)
                uids.append((source_uid, dest_uid))
        if not uids:
            copy_uid: Optional[CopyUid] = None
        else:
            copy_uid = CopyUid(dest.uid_validity, uids)
        return (copy_uid, await mbx.update_selected(selected))

    async def move_messages(self, selected: SelectedMailbox,
                            sequence_set: SequenceSet,
                            mailbox: str) \
            -> Tuple[Optional[CopyUid], SelectedMailbox]:
        mbx = await self._get_selected(selected)
        dest = await self._get_mailbox(mailbox, try_create=True)
        if dest.readonly:
            raise MailboxReadOnly(mailbox)
        dest_selected = self._pick_selected(selected, dest)
        uids: List[Tuple[int, int]] = []
        for _, source_uid in selected.messages.get_uids(sequence_set):
            dest_uid = await mbx.move(source_uid, dest,
                                      recent=not dest_selected)
            if dest_uid is not None:
                if dest_selected:
                    dest_selected.session_flags.add_recent(dest_uid)
                uids.append((source_uid, dest_uid))
        if not uids:
            copy_uid: Optional[CopyUid] = None
        else:
            copy_uid = CopyUid(dest.uid_validity, uids)
        return (copy_uid, await mbx.update_selected(selected))

    async def update_flags(self, selected: SelectedMailbox,
                           sequence_set: SequenceSet,
                           flag_set: FrozenSet[Flag],
                           mode: FlagOp = FlagOp.REPLACE) \
            -> Tuple[Iterable[Tuple[int, MessageT]], SelectedMailbox]:
        if selected.readonly:
            raise MailboxReadOnly()
        mbx = await self._get_selected(selected)
        permanent_flags = selected.permanent_flags & flag_set
        messages: List[Tuple[int, MessageT]] = []
        for seq, cached_msg in selected.messages.get_all(sequence_set):
            uid = cached_msg.uid
            msg = await mbx.update(uid, cached_msg, permanent_flags, mode)
            if not msg.expunged:
                selected.session_flags.update(uid, flag_set, mode)
            messages.append((seq, msg))
        return messages, await mbx.update_selected(selected)




from __future__ import annotations

import asyncio
from argparse import ArgumentParser
from asyncio import CancelledError
from ssl import SSLContext
from typing import Optional, Tuple, Sequence, Awaitable
from urllib.parse import urlparse

from grpclib.server import Server
from pymap.interfaces.backend import BackendInterface, ServiceInterface
from pymapadmin.local import get_admin_socket

from .handlers import handlers

__all__ = ['AdminService']


class AdminService(ServiceInterface):  # pragma: no cover
    """A pymap service implemented using a `grpc <https://grpc.io/>`_ server
    to perform admin functions on a running backend.

    This service can be exposed in *private* or *public* mode. Private mode
    does not use SSL/TLS encryption and authentication is not required, so it
    should be appropriately firewalled. Public mode requires both SSL/TLS
    encryption and the client provide credentials for the user being
    operated on, e.g. a user can change their own password but cannot change
    another user's password.

    The `pymap-admin <https://github.com/icgood/pymap-admin>`_ CLI may be used
    to interact with the admin service and serves as a reference client
    implementation.

    """

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('admin service')
        group.add_argument('--admin-private', metavar='HOST',
                           action='append', default=[],
                           help='private host:port to listen on')
        group.add_argument('--admin-public', metavar='HOST',
                           action='append', default=[],
                           help='public host:port to listen on')

    async def start(self) -> Awaitable:
        backend = self.backend
        private_hosts: Sequence[str] = self.config.args.admin_private
        public_hosts: Sequence[str] = self.config.args.admin_public
        ssl = self.config.ssl_context
        servers = [await self._start_local(backend)]
        servers += [await self._start(backend, False, host, None)
                    for host in private_hosts]
        servers += [await self._start(backend, True, host, ssl)
                    for host in public_hosts]
        return asyncio.create_task(self._run(servers))

    @classmethod
    def _new_server(cls, backend: BackendInterface, public: bool) -> Server:
        return Server([handler(backend, public) for _, handler in handlers])

    @classmethod
    async def _start_local(cls, backend: BackendInterface) -> Server:
        server = cls._new_server(backend, False)
        path = get_admin_socket(mkdir=True)
        await server.start(path=path)
        return server

    @classmethod
    async def _start(cls, backend: BackendInterface, public: bool,
                     host: str, ssl: Optional[SSLContext]) -> Server:
        server = cls._new_server(backend, public)
        host, port = cls._parse(host)
        await server.start(host=host, port=port, ssl=ssl, reuse_address=True)
        return server

    @classmethod
    def _parse(cls, host: str) -> Tuple[str, Optional[int]]:
        parsed = urlparse(host)
        if parsed.hostname:
            return parsed.hostname, parsed.port
        parsed = urlparse(f'//{host}')
        if parsed.hostname:
            return parsed.hostname, parsed.port
        raise ValueError(host)

    @classmethod
    async def _run(cls, servers: Sequence[Server]) -> None:
        try:
            for server in servers:
                await server.wait_closed()
        except CancelledError:
            for server in servers:
                server.close()
                await server.wait_closed()




from __future__ import annotations

from abc import ABCMeta
from typing import ClassVar, Optional, Tuple, Sequence, Iterable, List, \
    FrozenSet

from . import CommandAuth
from .. import Space, EndLine, Params
from ..exceptions import NotParseable, UnexpectedType, InvalidContent
from ..message import AppendMessage
from ..modutf7 import modutf7_decode
from ..primitives import ListP, String, LiteralString
from ..specials import Mailbox, DateTime, Flag, StatusAttribute, \
    ExtensionOption, ExtensionOptions
from ...bytes import rev

__all__ = ['AppendCommand', 'CreateCommand', 'DeleteCommand', 'ExamineCommand',
           'ListCommand', 'LSubCommand', 'RenameCommand', 'SelectCommand',
           'StatusCommand', 'SubscribeCommand', 'UnsubscribeCommand']


class CommandMailboxArg(CommandAuth, metaclass=ABCMeta):

    def __init__(self, tag: bytes, mailbox: Mailbox) -> None:
        super().__init__(tag)
        self.mailbox_obj = mailbox

    @property
    def mailbox(self) -> str:
        return str(self.mailbox_obj)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[CommandMailboxArg, memoryview]:
        _, buf = Space.parse(buf, params)
        mailbox, buf = Mailbox.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, mailbox), buf


class AppendCommand(CommandAuth):
    """The ``APPEND`` command adds a new message to a mailbox.

    See Also:
        `RFC 3502 6.3.11.
        <https://tools.ietf.org/html/rfc3502#section-6.3.11>`_

    Args:
        tag: The command tag.
        mailbox: The mailbox name.
        messages: List of tuples containing the raw messages bytes, the
            flags, and the internal timestamp to assign to the message.
        cancelled: True if the append command was cancelled.

    """

    command = b'APPEND'

    def __init__(self, tag: bytes, mailbox: Mailbox,
                 messages: Iterable[AppendMessage],
                 cancelled: bool = False, error: Exception = None) -> None:
        super().__init__(tag)
        self.mailbox_obj = mailbox
        self.messages: Sequence[AppendMessage] = list(messages)
        self.cancelled = cancelled
        self.error = error

    @property
    def mailbox(self) -> str:
        return str(self.mailbox_obj)

    @classmethod
    def _parse_msg(cls, name: str, buf: memoryview, params: Params) \
            -> Tuple[Optional[AppendMessage], memoryview]:
        _, buf = Space.parse(buf, params)
        try:
            params_copy = params.copy(list_expected=[Flag])
            flag_list, buf = ListP.parse(buf, params_copy)
        except UnexpectedType:
            raise
        except NotParseable:
            flags: FrozenSet[Flag] = frozenset()
        else:
            flags = frozenset(flag_list.get_as(Flag))
            _, buf = Space.parse(buf, params)
        try:
            date_time_p, buf = DateTime.parse(buf, params)
        except InvalidContent:
            raise
        except NotParseable:
            date_time = None
        else:
            date_time = date_time_p.value
            _, buf = Space.parse(buf, params)
        options_list: List[ExtensionOption] = []
        while True:
            try:
                option, buf = ExtensionOption.parse(buf, params)
            except NotParseable:
                break
            else:
                options_list.append(option)
        options = ExtensionOptions(options_list)
        try:
            message, buf = LiteralString.parse(buf, params)
        except NotParseable as exc:
            if options:
                literal = b''
            else:
                raise exc
        else:
            literal = message.value
            if literal == b'':
                return None, buf
        append_msg = AppendMessage(literal, date_time, flags, options)
        return append_msg, buf

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[AppendCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        mailbox, buf = Mailbox.parse(buf, params)
        messages: List[AppendMessage] = []
        error: Optional[Exception] = None
        cancelled = False
        while True:
            try:
                next_msg, buf = cls._parse_msg(mailbox.value, buf, params)
            except NotParseable:
                if not messages:
                    raise
                break
            else:
                if next_msg is None:
                    cancelled = True
                    break
                messages.append(next_msg)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, mailbox, messages, cancelled, error), buf


class CreateCommand(CommandMailboxArg):
    """The ``CREATE`` command creates a new mailbox."""

    command = b'CREATE'

    def __init__(self, tag: bytes, mailbox: Mailbox,
                 options: ExtensionOptions) -> None:
        super().__init__(tag, mailbox)
        self.options = options

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[CreateCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        mailbox, buf = Mailbox.parse(buf, params)
        options, buf = ExtensionOptions.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, mailbox, options), buf


class DeleteCommand(CommandMailboxArg):
    """The ``DELETE`` command deletes a mailbox."""

    command = b'DELETE'


class ListCommand(CommandAuth):
    """The ``LIST`` command lists existing mailboxes.

    Args:
        tag: The command tag.
        ref_name: The mailbox reference name.
        filter_: The mailbox filter string.

    """

    command = b'LIST'

    #: All mailboxes may be listed, not only subscribed mailboxes.
    only_subscribed: ClassVar[bool] = False

    _list_mailbox_pattern = rev.compile(br'[\x21\x23-\x27\x2A-\x5B'
                                        br'\x5D-\x7A\x7C\x7E]+')

    def __init__(self, tag: bytes, ref_name: str, filter_: str) -> None:
        super().__init__(tag)
        self.ref_name = ref_name
        self.filter = filter_

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[ListCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        ref_name, buf = Mailbox.parse(buf, params)
        _, buf = Space.parse(buf, params)
        match = cls._list_mailbox_pattern.match(buf)
        if match:
            filter_raw = match.group(0)
            buf = buf[match.end(0):]
            filter_ = modutf7_decode(filter_raw)
        else:
            filter_str, buf = String.parse(buf, params)
            filter_ = modutf7_decode(filter_str.value)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, ref_name.value, filter_), buf


class LSubCommand(ListCommand):
    """The ``LSUB`` command lists subscribed mailboxes."""

    command = b'LSUB'
    delegate = ListCommand

    #: Only subscribed mailboxes may be listed.
    only_subscribed: ClassVar[bool] = True


class RenameCommand(CommandAuth):
    """The ``RENAME`` command renames an existing mailbox.

    Args:
        tag: The command tag.
        from_mailbox: The existing mailbox name.
        to_mailbox: The desired mailbox name.

    """

    command = b'RENAME'

    def __init__(self, tag: bytes, from_mailbox: Mailbox,
                 to_mailbox: Mailbox, options: ExtensionOptions) -> None:
        super().__init__(tag)
        self.from_mailbox_obj = from_mailbox
        self.to_mailbox_obj = to_mailbox
        self.options = options

    @property
    def from_mailbox(self) -> str:
        return str(self.from_mailbox_obj)

    @property
    def to_mailbox(self) -> str:
        return str(self.to_mailbox_obj)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[RenameCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        from_mailbox, buf = Mailbox.parse(buf, params)
        _, buf = Space.parse(buf, params)
        to_mailbox, buf = Mailbox.parse(buf, params)
        options, buf = ExtensionOptions.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, from_mailbox, to_mailbox, options), buf


class SelectCommand(CommandMailboxArg):
    """The ``SELECT`` command selects a mailbox for querying, updates, and
    state changes.

    """

    command = b'SELECT'

    #: The mailbox will not be opened read-only unless the backend indicates
    #: that it must be.
    readonly: ClassVar[bool] = False

    def __init__(self, tag: bytes, mailbox: Mailbox,
                 options: ExtensionOptions) -> None:
        super().__init__(tag, mailbox)
        self.options = options

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[SelectCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        mailbox, buf = Mailbox.parse(buf, params)
        options, buf = ExtensionOptions.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, mailbox, options), buf


class ExamineCommand(SelectCommand):
    """The ``EXAMINE`` command selects a mailbox as read-only for querying and
    state changes.

    """

    command = b'EXAMINE'
    delegate = SelectCommand

    #: The mailbox will be opened read-only.
    readonly: ClassVar[bool] = True


class StatusCommand(CommandMailboxArg):
    """The ``STATUS`` command returns information about a mailbox without
    selecting it.

    Args:
        tag: The command tag.
        mailbox: The mailbox name.
        status_list: The status attributes to return.

    """

    command = b'STATUS'

    def __init__(self, tag: bytes, mailbox: Mailbox,
                 status_list: Sequence[StatusAttribute]) -> None:
        super().__init__(tag, mailbox)
        self.status_list = status_list

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[StatusCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        mailbox, buf = Mailbox.parse(buf, params)
        _, buf = Space.parse(buf, params)
        params_copy = params.copy(list_expected=[StatusAttribute])
        status_list_p, after = ListP.parse(buf, params_copy)
        if not status_list_p.value:
            raise NotParseable(buf)
        _, buf = EndLine.parse(after, params)
        status_list = status_list_p.get_as(StatusAttribute)
        return cls(params.tag, mailbox, status_list), buf


class SubscribeCommand(CommandMailboxArg):
    """The ``SUBSCRIBE`` command subscribes to an existing mailbox."""

    command = b'SUBSCRIBE'


class UnsubscribeCommand(CommandMailboxArg):
    """The ``UNSUBSCRIBE`` command unsubscribes from a subscribed mailbox."""

    command = b'UNSUBSCRIBE'



# based on: https://github.com/python/typeshed/blob/master/stdlib/3/re.pyi

from re import RegexFlag
from typing import (
    List, Iterator, overload, Callable, Tuple,
    Match, Pattern, Any, Optional, Union
)

_FlagsType = Union[int, RegexFlag]
_Str = Union[bytes, memoryview]

class _RevPattern:
    def search(self, string: _Str, pos: int = ..., endpos: int = ...) -> Optional[Match[bytes]]: ...
    def match(self, string: _Str, pos: int = ..., endpos: int = ...) -> Optional[Match[bytes]]: ...
    def fullmatch(self, string: _Str, pos: int = ..., endpos: int = ...) -> Optional[Match[bytes]]: ...
    def split(self, string: _Str, maxsplit: int = ...) -> List[bytes]: ...
    def findall(self, string: _Str, pos: int = ..., endpos: int = ...) -> List[Any]: ...
    def finditer(self, string: _Str, pos: int = ..., endpos: int = ...) -> Iterator[Match[bytes]]: ...
    @overload
    def sub(self, repl: bytes, string: _Str, count: int = ...) -> bytes: ...
    @overload
    def sub(self, repl: Callable[[Match[bytes]], bytes], string: _Str, count: int = ...) -> bytes: ...
    @overload
    def subn(self, repl: bytes, string: _Str, count: int = ...) -> Tuple[bytes, int]: ...
    @overload
    def subn(self, repl: Callable[[Match[bytes]], bytes], string: _Str, count: int = ...) -> Tuple[bytes, int]: ...

@overload
def compile(pattern: bytes, flags: _FlagsType = ...) -> _RevPattern: ...
@overload
def compile(pattern: Pattern[bytes], flags: _FlagsType = ...) -> _RevPattern: ...



"""Utilities for managing IMAP searches."""

from __future__ import annotations

import re
from abc import abstractmethod, ABCMeta
from datetime import datetime
from typing import AnyStr, FrozenSet, Optional, Iterable
from typing_extensions import Final

from .exceptions import SearchNotAllowed
from .interfaces.message import MessageInterface, LoadedMessageInterface
from .parsing.specials import ObjectId, SearchKey, SequenceSet
from .parsing.specials.flag import Flag, Answered, Deleted, Draft, Flagged, \
    Recent, Seen
from .selected import SelectedMailbox

__all__ = ['SearchParams', 'SearchCriteria', 'SearchCriteriaSet']


class SearchParams:
    """Defines certain parameters and routines necessary to process any
    kind of search criteria. If a parameter is not supplied, or a method not
    implemented, any search keys that require it will fail.

    Args:
        selected: The active mailbox session.
        max_seq: The highest message sequence ID in the mailbox.
        max_uid: The highest message UID in the mailbox.
        disabled: Search keys that should be disabled.

    """

    __slots__ = ['selected', 'disabled', 'max_seq', 'max_uid', 'session_flags']

    def __init__(self, selected: SelectedMailbox, *,
                 disabled: Iterable[bytes] = None) -> None:
        self.selected: Final = selected
        self.disabled: Final = frozenset(disabled or [])
        self.max_seq: Final = selected.messages.exists
        self.max_uid: Final = selected.messages.max_uid
        self.session_flags: Final = selected.session_flags


class SearchCriteria(metaclass=ABCMeta):
    """Base class for different types of search criteria.

    Args:
        params: The parameters that may be used by some searches.

    """

    def __init__(self, params: SearchParams) -> None:
        self.params = params

    @classmethod
    def _in(cls, substr: AnyStr, data: AnyStr) -> bool:
        re_flags = re.I | re.A
        escaped_substr = re.escape(substr)
        return re.search(escaped_substr, data, re_flags) is not None

    @abstractmethod
    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        """Implemented by sub-classes to define the search criteria.

        Args:
            msg_seq: The message sequence ID.
            msg: The message object.
            loaded_msg: The message content.

        """
        ...

    @classmethod
    def of(cls, key: SearchKey, params: SearchParams) -> SearchCriteria:
        """Factory method for producing a search criteria sub-class from a
        search key.

        Args:
            key: The search key defining the criteria.
            params: The parameters that may be used by some searches.

        """
        key_name = key.value
        if key_name in params.disabled:
            raise SearchNotAllowed(key)
        elif key.inverse:
            return InverseSearchCriteria(key.not_inverse, params)
        elif key_name == b'SEQSET':
            return SequenceSetSearchCriteria(key.filter_sequence_set, params)
        elif key_name == b'KEYSET':
            return SearchCriteriaSet(key.filter_key_set, params)
        elif key_name == b'ALL':
            return AllSearchCriteria(params)
        elif key_name == b'OR':
            left_key, right_key = key.filter_key_or
            return OrSearchCriteria(left_key, right_key, params)
        elif key_name == b'EMAILID':
            return HasEmailIdSearchCriteria(key.filter_object_id, params)
        elif key_name == b'THREADID':
            return HasThreadIdSearchCriteria(key.filter_object_id, params)
        elif key_name == b'ANSWERED':
            return HasFlagSearchCriteria(Answered, True, params)
        elif key_name == b'UNANSWERED':
            return HasFlagSearchCriteria(Answered, False, params)
        elif key_name == b'DELETED':
            return HasFlagSearchCriteria(Deleted, True, params)
        elif key_name == b'UNDELETED':
            return HasFlagSearchCriteria(Deleted, False, params)
        elif key_name == b'DRAFT':
            return HasFlagSearchCriteria(Draft, True, params)
        elif key_name == b'UNDRAFT':
            return HasFlagSearchCriteria(Draft, False, params)
        elif key_name == b'FLAGGED':
            return HasFlagSearchCriteria(Flagged, True, params)
        elif key_name == b'UNFLAGGED':
            return HasFlagSearchCriteria(Flagged, False, params)
        elif key_name == b'RECENT':
            return HasFlagSearchCriteria(Recent, True, params)
        elif key_name == b'OLD':
            return HasFlagSearchCriteria(Recent, False, params)
        elif key_name == b'SEEN':
            return HasFlagSearchCriteria(Seen, True, params)
        elif key_name == b'UNSEEN':
            return HasFlagSearchCriteria(Seen, False, params)
        elif key_name == b'KEYWORD':
            return HasFlagSearchCriteria(key.filter_flag, True, params)
        elif key_name == b'UNKEYWORD':
            return HasFlagSearchCriteria(key.filter_flag, False, params)
        elif key_name == b'NEW':
            return NewSearchCriteria(params)
        elif key_name == b'BEFORE':
            return DateSearchCriteria(key.filter_datetime, '<', params)
        elif key_name == b'ON':
            return DateSearchCriteria(key.filter_datetime, '=', params)
        elif key_name == b'SINCE':
            return DateSearchCriteria(key.filter_datetime, '>=', params)
        elif key_name == b'SENTBEFORE':
            return HeaderDateSearchCriteria(key.filter_datetime, '<', params)
        elif key_name == b'SENTON':
            return HeaderDateSearchCriteria(key.filter_datetime, '=', params)
        elif key_name == b'SENTSINCE':
            return HeaderDateSearchCriteria(key.filter_datetime, '>=', params)
        elif key_name == b'SMALLER':
            return SizeSearchCriteria(key.filter_int, '<', params)
        elif key_name == b'LARGER':
            return SizeSearchCriteria(key.filter_int, '>', params)
        elif key_name in (b'BCC', b'CC', b'FROM', b'SUBJECT', b'TO'):
            return EnvelopeSearchCriteria(key_name, key.filter_str, params)
        elif key_name == b'HEADER':
            name, value = key.filter_header
            return HeaderSearchCriteria(name, value, params)
        elif key_name in (b'BODY', b'TEXT'):
            return BodySearchCriteria(key.filter_str, params)
        raise SearchNotAllowed(key)


class SearchCriteriaSet(SearchCriteria):
    """Search criteria composed of a set of search criteria that must all
    match. If the set is empty, nothing will match.

    Args:
        keys: The set of search keys that must match.
        params: The parameters that may be used by some searches.

    """

    def __init__(self, keys: FrozenSet[SearchKey],
                 params: SearchParams) -> None:
        super().__init__(params)
        self.all_criteria = [SearchCriteria.of(key, params) for key in keys]

    @property
    def sequence_set(self) -> SequenceSet:
        """The sequence set to use when finding the messages to match against.
        This will default to all messages unless the search criteria set
        contains a sequence set.

        """
        try:
            seqset_crit = next(crit for crit in self.all_criteria
                               if isinstance(crit, SequenceSetSearchCriteria))
        except StopIteration:
            return SequenceSet.all()
        else:
            return seqset_crit.seq_set

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        """The message matches if all the defined search key criteria match.

        Args:
            msg_seq: The message sequence ID.
            msg: The message object.

        """
        return all(crit.matches(msg_seq, msg, loaded_msg)
                   for crit in self.all_criteria)


class InverseSearchCriteria(SearchCriteria):
    """Matches only if the given search criteria does not match."""

    def __init__(self, key: SearchKey, params: SearchParams) -> None:
        super().__init__(params)
        self.key = SearchCriteria.of(key, params)

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return not self.key.matches(msg_seq, msg, loaded_msg)


class AllSearchCriteria(SearchCriteria):
    """Always matches anything."""

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return True


class OrSearchCriteria(SearchCriteria):
    """Matches if either of the search criteria match."""

    def __init__(self, left: SearchKey, right: SearchKey,
                 params: SearchParams) -> None:
        super().__init__(params)
        self.left = SearchCriteria.of(left, self.params)
        self.right = SearchCriteria.of(right, self.params)

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return (self.left.matches(msg_seq, msg, loaded_msg)
                or self.right.matches(msg_seq, msg, loaded_msg))


class SequenceSetSearchCriteria(SearchCriteria):
    """Matches if the message is contained in the sequence set."""

    def __init__(self, seq_set: SequenceSet, params: SearchParams) -> None:
        super().__init__(params)
        self.seq_set = seq_set
        if seq_set.uid:
            self.flat = seq_set.flatten(params.max_uid)
        else:
            self.flat = seq_set.flatten(params.max_seq)

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        if self.seq_set.uid:
            return msg.uid in self.flat
        else:
            return msg_seq in self.flat


class HasEmailIdSearchCriteria(SearchCriteria):
    """Matches if the message has the same
    :attr:`~pymap.interfaces.message.MessageInterface.email_id` value.

    """

    def __init__(self, email_id: ObjectId, params: SearchParams) -> None:
        super().__init__(params)
        self.email_id = email_id

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        try:
            email_id = msg.email_id
        except ValueError:
            return False
        else:
            return self.email_id == email_id


class HasThreadIdSearchCriteria(SearchCriteria):
    """Matches if the message has the same
    :attr:`~pymap.interfaces.message.MessageInterface.thread_id` value.

    """

    def __init__(self, thread_id: ObjectId, params: SearchParams) -> None:
        super().__init__(params)
        self.thread_id = thread_id

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return self.thread_id == msg.thread_id


class HasFlagSearchCriteria(SearchCriteria):
    """Matches if the message has the given flag in their permanent or
    session flag sets.

    """

    def __init__(self, flag: Flag, expected: bool,
                 params: SearchParams) -> None:
        super().__init__(params)
        self.flag = flag
        self.expected = expected

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        has_flag = self.flag in msg.get_flags(self.params.session_flags)
        expected = self.expected
        return (has_flag and expected) or (not expected and not has_flag)


class NewSearchCriteria(SearchCriteria):
    """Matches if the message is considered "new", i.e. recent and unseen."""

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        flags = msg.get_flags(self.params.session_flags)
        return Recent in flags and Seen not in flags


class DateSearchCriteria(SearchCriteria):
    """Matches by comparing against the internal date of the message."""

    def __init__(self, when: datetime, op: str, params: SearchParams) -> None:
        super().__init__(params)
        self.when = when.date()
        self.op = op

    @classmethod
    def _get_msg_date(cls, msg: MessageInterface,
                      loaded_msg: LoadedMessageInterface) \
            -> Optional[datetime]:
        return msg.internal_date

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        msg_datetime = self._get_msg_date(msg, loaded_msg)
        if msg_datetime is None:
            return False
        msg_date = msg_datetime.date()
        if self.op == '<':  # BEFORE
            return msg_date < self.when
        elif self.op == '=':  # ON
            return msg_date == self.when
        elif self.op == '>=':  # SINCE
            return msg_date >= self.when
        raise ValueError(self.op)


class HeaderDateSearchCriteria(DateSearchCriteria):
    """Matches by comparing against the ``Date:`` header of the message."""

    @classmethod
    def _get_msg_date(cls, msg: MessageInterface,
                      loaded_msg: LoadedMessageInterface) \
            -> Optional[datetime]:
        envelope = loaded_msg.get_envelope_structure()
        return envelope.date.datetime if envelope.date else None


class SizeSearchCriteria(SearchCriteria):
    """Matches by comparing against the size of the message."""

    def __init__(self, size: int, op: str, params: SearchParams) -> None:
        super().__init__(params)
        self.size = size
        self.op = op

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        size = loaded_msg.get_size()
        if self.op == '<':
            return size < self.size
        elif self.op == '>':
            return size > self.size
        raise ValueError(self.op)


class EnvelopeSearchCriteria(SearchCriteria):
    """Matches by checking for strings within various fields of the envelope
    structure.

    """

    def __init__(self, key: bytes, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.key = key
        self.value = value

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        envelope = loaded_msg.get_envelope_structure()
        if self.key == b'BCC':
            if not envelope.bcc:
                return False
            return any(self._in(self.value, str(bcc)) for bcc in envelope.bcc)
        elif self.key == b'CC':
            if not envelope.cc:
                return False
            return any(self._in(self.value, str(cc)) for cc in envelope.cc)
        elif self.key == b'FROM':
            if not envelope.from_:
                return False
            return any(self._in(self.value, str(from_))
                       for from_ in envelope.from_)
        elif self.key == b'SUBJECT':
            if not envelope.subject:
                return False
            return self._in(self.value, str(envelope.subject))
        elif self.key == b'TO':
            if not envelope.to:
                return False
            return any(self._in(self.value, str(to)) for to in envelope.to)
        raise ValueError(self.key)


class HeaderSearchCriteria(SearchCriteria):
    """Matches if the message has a header containing a value."""

    def __init__(self, name: str, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.name = name.encode('ascii')
        self.value = value

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        values = loaded_msg.get_header(self.name)
        return any(self._in(self.value, value) for value in values)


class BodySearchCriteria(SearchCriteria):
    """Matches if the message body contains a value."""

    def __init__(self, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.value = bytes(value, 'utf-8', 'replace')

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return loaded_msg.contains(self.value)




import unittest

from pymap.flags import FlagOp
from pymap.parsing import Params
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.command.select import ExpungeCommand, CopyCommand, \
    MoveCommand, FetchCommand, StoreCommand, SearchCommand, \
    UidExpungeCommand, UidCopyCommand, UidMoveCommand, UidFetchCommand, \
    UidStoreCommand, UidSearchCommand, IdleCommand
from pymap.parsing.specials import FetchAttribute, SearchKey, Flag


class TestExpungeCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = ExpungeCommand.parse(b' \n  ', Params())
        self.assertFalse(ret.uid)
        self.assertEqual(b'  ', buf)

    def test_parse_uid(self):
        ret, buf = UidExpungeCommand.parse(b' 12,13,14 \n  ', Params())
        self.assertTrue(ret.uid)
        self.assertTrue(ret.uid_set.uid)
        self.assertEqual([12, 13, 14], ret.uid_set.value)
        self.assertEqual(b'  ', buf)


class TestCopyCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = CopyCommand.parse(b' 1,2,3 mbx\n  ', Params())
        self.assertFalse(ret.uid)
        self.assertFalse(ret.sequence_set.uid)
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertEqual('mbx', ret.mailbox)
        self.assertEqual(b'  ', buf)

    def test_parse_uid(self):
        ret, buf = UidCopyCommand.parse(b' 1,2,3 mbx\n  ', Params())
        self.assertTrue(ret.uid)
        self.assertTrue(ret.sequence_set.uid)


class TestMoveCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = MoveCommand.parse(b' 1,2,3 mbx\n  ', Params())
        self.assertFalse(ret.uid)
        self.assertFalse(ret.sequence_set.uid)
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertEqual('mbx', ret.mailbox)
        self.assertEqual(b'  ', buf)

    def test_parse_uid(self):
        ret, buf = UidMoveCommand.parse(b' 1,2,3 mbx\n  ', Params())
        self.assertTrue(ret.uid)
        self.assertTrue(ret.sequence_set.uid)


class TestFetchCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = FetchCommand.parse(b' 1,2,3 ENVELOPE\n  ', Params())
        self.assertFalse(ret.uid)
        self.assertFalse(ret.sequence_set.uid)
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertEqual(1, len(ret.attributes))
        self.assertListEqual([FetchAttribute(b'ENVELOPE')], ret.attributes)
        self.assertEqual(b'  ', buf)

    def test_parse_uid(self):
        ret, buf = UidFetchCommand.parse(b' 1,2,3 ENVELOPE\n  ', Params())
        self.assertTrue(ret.uid)
        self.assertTrue(ret.sequence_set.uid)

    def test_parse_list(self):
        ret, buf = FetchCommand.parse(b' 1,2,3 (FLAGS ENVELOPE)\n  ', Params())
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertListEqual([FetchAttribute(b'FLAGS'),
                              FetchAttribute(b'ENVELOPE')], ret.attributes)
        self.assertEqual(b'  ', buf)

    def test_parse_macro_all(self):
        ret, buf = FetchCommand.parse(b' 1,2,3 ALL\n  ', Params())
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertListEqual([FetchAttribute(b'FLAGS'),
                              FetchAttribute(b'INTERNALDATE'),
                              FetchAttribute(b'RFC822.SIZE'),
                              FetchAttribute(b'ENVELOPE')], ret.attributes)
        self.assertEqual(b'  ', buf)

    def test_parse_macro_full(self):
        ret, buf = FetchCommand.parse(b' 1,2,3 FULL\n  ', Params())
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertListEqual([FetchAttribute(b'FLAGS'),
                              FetchAttribute(b'INTERNALDATE'),
                              FetchAttribute(b'RFC822.SIZE'),
                              FetchAttribute(b'ENVELOPE'),
                              FetchAttribute(b'BODY')], ret.attributes)
        self.assertEqual(b'  ', buf)

    def test_parse_macro_fast(self):
        ret, buf = FetchCommand.parse(b' 1,2,3 FAST\n  ', Params())
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertListEqual([FetchAttribute(b'FLAGS'),
                              FetchAttribute(b'INTERNALDATE'),
                              FetchAttribute(b'RFC822.SIZE')], ret.attributes)
        self.assertEqual(b'  ', buf)


class TestStoreCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = StoreCommand.parse(
            b' 1,2,3 +FLAGS.SILENT (\\Seen)\n  ', Params())
        self.assertFalse(ret.uid)
        self.assertFalse(ret.sequence_set.uid)
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertSetEqual({Flag(br'\Seen')}, ret.flag_set)
        self.assertEqual(FlagOp.ADD, ret.mode)
        self.assertTrue(ret.silent)
        self.assertEqual(b'  ', buf)

    def test_parse_uid(self):
        ret, buf = UidStoreCommand.parse(
            b' 1,2,3 +FLAGS.SILENT (\\Seen)\n  ', Params())
        self.assertTrue(ret.uid)
        self.assertTrue(ret.sequence_set.uid)

    def test_parse_simple(self):
        ret, buf = StoreCommand.parse(b' 1,2,3 FLAGS \\Seen\n  ', Params())
        self.assertEqual([1, 2, 3], ret.sequence_set.value)
        self.assertSetEqual({Flag(br'\Seen')}, ret.flag_set)
        self.assertEqual(FlagOp.REPLACE, ret.mode)
        self.assertFalse(ret.silent)
        self.assertEqual(b'  ', buf)

    def test_parse_error(self):
        with self.assertRaises(NotParseable):
            StoreCommand.parse(b' 1,2,3 TEST (\\Seen)\n', Params())


class TestSearchCommand(unittest.TestCase):

    def test_parse_charset(self):
        ret, buf = SearchCommand._parse_charset(
            b' CHARSET "utf-8"  ', Params())
        self.assertEqual('utf-8', ret)
        self.assertEqual(b'  ', buf)

    def test_parse_charset_error(self):
        with self.assertRaises(NotParseable):
            SearchCommand._parse_charset(b' CHARSET "test"', Params())

    def test_parse_charset_default(self):
        ret, buf = SearchCommand._parse_charset(b'  ', Params())
        self.assertIsNone(ret)
        self.assertEqual(b'  ', buf)

    def test_parse(self):
        ret, buf = SearchCommand.parse(b' ALL\n  ', Params())
        self.assertFalse(ret.uid)
        self.assertSetEqual({SearchKey(b'ALL')}, ret.keys)
        self.assertIsNone(ret.charset)
        self.assertEqual(b'  ', buf)

    def test_parse_uid(self):
        ret, buf = UidSearchCommand.parse(b' ALL\n  ', Params())
        self.assertTrue(ret.uid)

    def test_parse_error(self):
        with self.assertRaises(NotParseable):
            SearchCommand.parse(b' TEST\n', Params())


class TestIdleCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = IdleCommand.parse(b' \r\nabc', Params())
        self.assertIsInstance(ret, IdleCommand)
        self.assertEqual(b'DONE', ret.continuation)
        self.assertEqual(b'abc', buf)

    def test_parse_error(self):
        with self.assertRaises(NotParseable):
            IdleCommand.parse(b' STUFF\r\n', Params())

    def test_parse_done(self):
        ok, buf = IdleCommand(b'tag').parse_done(b'DONE\r\nabc')
        self.assertTrue(ok)
        self.assertEqual(b'abc', buf)

    def test_parse_done_bad(self):
        ok, buf = IdleCommand(b'tag').parse_done(b' DONE\r\n')
        self.assertFalse(ok)
        self.assertEqual(b'', buf)
        ok, buf = IdleCommand(b'tag').parse_done(b'TEST\r\n')
        self.assertFalse(ok)
        self.assertEqual(b'', buf)




from __future__ import annotations

from typing import IO, TypeVar, Type, Iterable, Union, Mapping, FrozenSet, \
    Sequence

from pymap.parsing.specials.flag import Flag, Seen, Flagged, Deleted, Draft, \
    Answered

from .io import FileReadable

__all__ = ['MaildirFlags']

_MFT = TypeVar('_MFT', bound='MaildirFlags')


class MaildirFlags(FileReadable):
    """Maintains a set of IMAP keywords (non-standard flags) that are available
    for use on messages. This uses a custom file format to define keywords,
    which might look like this::

        0 $Junk
        1 $NonJunk

    The lower-case letter codes that correspond to each keyword start with
    ``'a'`` for 0, ``'b'`` for 1, etc. and up to 26 are supported.

    Args:
        keywords: The list of keywords available for use on messages.

    See Also:
        `IMAP Keywords
        <https://wiki.dovecot.org/MailboxFormat/Maildir#line-40>`_

    """

    _from_sys: Mapping[Flag, str] = {Seen: 'S',
                                     Flagged: 'F',
                                     Deleted: 'T',
                                     Draft: 'D',
                                     Answered: 'R'}

    _to_sys: Mapping[str, Flag] = {'S': Seen,
                                   'F': Flagged,
                                   'T': Deleted,
                                   'D': Draft,
                                   'R': Answered}

    def __init__(self, keywords: Sequence[Flag]) -> None:
        super().__init__()
        if len(keywords) > 26:
            raise ValueError(keywords)
        self._keywords = frozenset(keywords)
        self._to_kwd = {chr(ord('a') + i): kwd
                        for i, kwd in enumerate(keywords)}
        self._from_kwd = {kwd: chr(ord('a') + i)
                          for i, kwd in enumerate(keywords)}

    @property
    def permanent_flags(self) -> FrozenSet[Flag]:
        """Return the set of all permanent flags, system and keyword."""
        return self.system_flags | self.keywords

    @property
    def system_flags(self) -> FrozenSet[Flag]:
        """Return the set of defined IMAP system flags."""
        return frozenset(self._from_sys.keys())

    @property
    def keywords(self) -> FrozenSet[Flag]:
        """Return the set of available IMAP keywords."""
        return self._keywords

    def to_maildir(self, flags: Iterable[Union[bytes, Flag]]) -> str:
        """Return the string of letter codes that are used to map to defined
        IMAP flags and keywords.

        Args:
            flags: The flags and keywords to map.

        """
        codes = []
        for flag in flags:
            if isinstance(flag, bytes):
                flag = Flag(flag)
            from_sys = self._from_sys.get(flag)
            if from_sys is not None:
                codes.append(from_sys)
            else:
                from_kwd = self._from_kwd.get(flag)
                if from_kwd is not None:
                    codes.append(from_kwd)
        return ''.join(codes)

    def from_maildir(self, codes: str) -> FrozenSet[Flag]:
        """Return the set of IMAP flags that correspond to the letter codes.

        Args:
            codes: The letter codes to map.

        """
        flags = set()
        for code in codes:
            if code == ',':
                break
            to_sys = self._to_sys.get(code)
            if to_sys is not None:
                flags.add(to_sys)
            else:
                to_kwd = self._to_kwd.get(code)
                if to_kwd is not None:
                    flags.add(to_kwd)
        return frozenset(flags)

    @classmethod
    def get_file(cls) -> str:
        return 'dovecot-keywords'

    @classmethod
    def get_default(cls: Type[_MFT], base_dir: str) -> _MFT:
        return cls([])

    @classmethod
    def open(cls: Type[_MFT], base_dir: str, fp: IO[str]) -> _MFT:
        ret = []
        for line in fp:
            i, kwd = line.split()
            if kwd.startswith('\\'):
                raise ValueError(kwd)
            ret.append((i, kwd))
        return cls([Flag(kwd) for _, kwd in sorted(ret)])




from __future__ import annotations

import hashlib
from bisect import bisect_left
from collections import OrderedDict
from datetime import datetime
from itertools import islice
from typing import Tuple, Sequence, Dict, Optional, Iterable, \
    AsyncIterable, List, Set, AbstractSet, FrozenSet
from weakref import finalize, WeakKeyDictionary, WeakValueDictionary

from pymap.bytes import HashStream
from pymap.concurrent import Event, ReadWriteLock
from pymap.context import subsystem
from pymap.flags import FlagOp
from pymap.interfaces.message import CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.message import BaseMessage, BaseLoadedMessage
from pymap.mime import MessageContent
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials import ObjectId, FetchRequirement
from pymap.parsing.specials.flag import Flag, Seen
from pymap.selected import SelectedSet, SelectedMailbox
from pymap.threads import ThreadKey

from ..mailbox import MailboxDataInterface, MailboxSetInterface

__all__ = ['Message', 'MailboxData', 'MailboxSet']


class Message(BaseMessage):

    __slots__ = ['_recent', '_content']

    def __init__(self, uid: int, internal_date: datetime,
                 permanent_flags: Iterable[Flag], *, expunged: bool = False,
                 email_id: ObjectId = None, thread_id: ObjectId = None,
                 recent: bool = False, content: MessageContent = None) -> None:
        super().__init__(uid, internal_date, permanent_flags,
                         expunged=expunged, email_id=email_id,
                         thread_id=thread_id)
        self._recent = recent
        self._content = content

    @classmethod
    def copy(cls, msg: Message, *, uid: int = None, recent: bool = False,
             expunged: bool = False) -> Message:
        if uid is None:
            uid = msg.uid
        return cls(uid, msg.internal_date, msg.permanent_flags,
                   expunged=expunged, email_id=msg.email_id,
                   thread_id=msg.thread_id, recent=recent,
                   content=msg._content)

    @property
    def recent(self) -> bool:
        return self._recent

    @recent.setter
    def recent(self, recent: bool) -> None:
        self._recent = recent

    async def load_content(self, requirement: FetchRequirement) \
            -> LoadedMessage:
        return LoadedMessage(self, requirement, self._content)


class LoadedMessage(BaseLoadedMessage):
    pass


class _ModSequenceMapping:

    def __init__(self) -> None:
        super().__init__()
        self._highest = 0
        self._uids: Dict[int, int] = {}
        self._updates: Dict[int, Set[int]] = {}
        self._expunges: Dict[int, Set[int]] = {}
        self._mod_seqs_order: List[int] = []

    @property
    def highest(self) -> int:
        return self._highest

    def _remove_prev(self, uid: int, prev_mod_seq: int,
                     data: Dict[int, Set[int]]) -> None:
        uid_set = data.get(prev_mod_seq, None)
        if uid_set is not None:
            uid_set.discard(uid)
            if not uid_set:
                del data[prev_mod_seq]
                self._mod_seqs_order.remove(prev_mod_seq)

    def _set(self, uids: Iterable[int], data: Dict[int, Set[int]]) -> None:
        self._highest = mod_seq = self._highest + 1
        self._mod_seqs_order.append(mod_seq)
        new_uid_set = data.setdefault(mod_seq, set())
        new_uid_set.update(uids)
        for uid in uids:
            prev_mod_seq = self._uids.get(uid, None)
            self._uids[uid] = mod_seq
            if prev_mod_seq is not None:
                self._remove_prev(uid, prev_mod_seq, self._updates)
                self._remove_prev(uid, prev_mod_seq, self._expunges)

    def update(self, uids: Iterable[int]) -> None:
        return self._set(uids, self._updates)

    def expunge(self, uids: Iterable[int]) -> None:
        return self._set(uids, self._expunges)

    def find_updated(self, mod_seq: int) \
            -> Tuple[AbstractSet[int], AbstractSet[int]]:
        updates_ret: Set[int] = set()
        expunges_ret: Set[int] = set()
        updates = self._updates
        expunges = self._expunges
        mod_seqs_order = self._mod_seqs_order
        mod_seqs_len = len(mod_seqs_order)
        idx = bisect_left(mod_seqs_order, mod_seq, 0, mod_seqs_len)
        for newer_mod_seq in islice(mod_seqs_order, idx, mod_seqs_len):
            updates_set = updates.get(newer_mod_seq)
            expunges_set = expunges.get(newer_mod_seq)
            if updates_set is not None:
                updates_ret.update(updates_set)
            if expunges_set is not None:
                expunges_ret.update(expunges_set)
        return updates_ret, expunges_ret


class _ContentCache:

    def __init__(self) -> None:
        super().__init__()
        self._email_ids: Dict[bytes, ObjectId] = {}
        self._hashes: Dict[ObjectId, bytes] = {}
        self._content: WeakValueDictionary[ObjectId, MessageContent] = \
            WeakValueDictionary()

    def add(self, content: MessageContent) -> ObjectId:
        msg_hash = HashStream(hashlib.sha1()).digest(content)
        existing = self._email_ids.get(msg_hash)
        if existing is not None:
            return existing
        email_id = ObjectId.random_email_id()
        self._email_ids[msg_hash] = email_id
        self._hashes[email_id] = msg_hash
        self._content[email_id] = content
        finalize(content, self._remove, msg_hash, email_id)
        return email_id

    def _remove(self, msg_hash: bytes, email_id: ObjectId) -> None:
        self._email_ids.pop(msg_hash, None)
        self._hashes.pop(email_id, None)

    def get(self, email_id: ObjectId) -> MessageContent:
        return self._content[email_id]


class _ThreadCache:

    def __init__(self) -> None:
        super().__init__()
        self._thread_ids: WeakKeyDictionary[ThreadKey, ObjectId] = \
            WeakKeyDictionary()
        self._ref: WeakKeyDictionary[MessageContent, Sequence[ThreadKey]] = \
            WeakKeyDictionary()

    def add(self, content: MessageContent) -> ObjectId:
        self._ref[content] = thread_keys = ThreadKey.get_all(content.header)
        for thread_key in thread_keys:
            thread_id = self._thread_ids.get(thread_key)
            if thread_id is not None:
                break
        else:
            thread_id = ObjectId.random_thread_id()
        for thread_key in thread_keys:
            self._thread_ids.setdefault(thread_key, thread_id)
        return thread_id


class MailboxData(MailboxDataInterface[Message]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxDataInterface`
    for the dict backend.

    """

    def __init__(self, content_cache: _ContentCache,
                 thread_cache: _ThreadCache) -> None:
        self._mailbox_id = ObjectId.random_mailbox_id()
        self._content_cache = content_cache
        self._thread_cache = thread_cache
        self._readonly = False
        self._updated = subsystem.get().new_event()
        self._messages_lock = subsystem.get().new_rwlock()
        self._selected_set = SelectedSet()
        self._uid_validity = MailboxSnapshot.new_uid_validity()
        self._max_uid = 100
        self._mod_sequences = _ModSequenceMapping()
        self._messages: Dict[int, Message] = OrderedDict()

    @property
    def mailbox_id(self) -> ObjectId:
        return self._mailbox_id

    @property
    def readonly(self) -> bool:
        return self._readonly

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def messages_lock(self) -> ReadWriteLock:
        return self._messages_lock

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    async def update_selected(self, selected: SelectedMailbox, *,
                              wait_on: Event = None) -> SelectedMailbox:
        if wait_on is not None:
            either_event = wait_on.or_event(self._updated)
            await either_event.wait()
        mod_sequence = selected.mod_sequence
        selected.mod_sequence = self._mod_sequences.highest
        if mod_sequence is None:
            all_messages = list(self._messages.values())
            selected.add_updates(all_messages, [])
        else:
            updated, expunged = self._mod_sequences.find_updated(mod_sequence)
            updated_messages = [self._messages[uid] for uid in updated
                                if uid in self._messages]
            selected.add_updates(updated_messages, expunged)
        return selected

    async def append(self, append_msg: AppendMessage, *,
                     recent: bool = False) -> Message:
        when = append_msg.when or datetime.now()
        content = MessageContent.parse(append_msg.literal)
        email_id = self._content_cache.add(content)
        thread_id = self._thread_cache.add(content)
        async with self.messages_lock.write_lock():
            self._max_uid = new_uid = self._max_uid + 1
            message = Message(new_uid, when, append_msg.flag_set,
                              email_id=email_id, thread_id=thread_id,
                              recent=recent, content=content)
            self._messages[new_uid] = message
            self._mod_sequences.update([new_uid])
            self._updated.set()
            return message

    async def copy(self, uid: int, destination: MailboxData, *,
                   recent: bool = False) -> Optional[int]:
        async with self.messages_lock.read_lock():
            try:
                message = self._messages[uid]
            except KeyError:
                return None
        async with destination.messages_lock.write_lock():
            destination._max_uid = dest_uid = destination._max_uid + 1
            new_msg = Message.copy(message, uid=dest_uid, recent=recent)
            destination._messages[dest_uid] = new_msg
            destination._mod_sequences.update([dest_uid])
            destination._updated.set()
        return dest_uid

    async def move(self, uid: int, destination: MailboxData, *,
                   recent: bool = False) -> Optional[int]:
        async with self.messages_lock.write_lock():
            try:
                message = self._messages.pop(uid)
            except KeyError:
                return None
            self._mod_sequences.expunge([uid])
            self._updated.set()
        async with destination.messages_lock.write_lock():
            destination._max_uid = dest_uid = destination._max_uid + 1
            new_msg = Message.copy(message, uid=dest_uid, recent=recent)
            destination._messages[dest_uid] = new_msg
            destination._mod_sequences.update([dest_uid])
            destination._updated.set()
        return dest_uid

    async def get(self, uid: int, cached_msg: CachedMessage) -> Message:
        if uid < 1 or uid > self._max_uid:
            raise IndexError(uid)
        async with self.messages_lock.read_lock():
            msg = self._messages.get(uid)
        if msg is None:
            if not isinstance(cached_msg, Message):
                raise TypeError(cached_msg)
            msg = Message.copy(cached_msg, expunged=True)
        return msg

    async def update(self, uid: int, cached_msg: CachedMessage,
                     flag_set: FrozenSet[Flag], mode: FlagOp) -> Message:
        msg = await self.get(uid, cached_msg)
        msg.permanent_flags = mode.apply(msg.permanent_flags, flag_set)
        self._mod_sequences.update([uid])
        self._updated.set()
        return msg

    async def delete(self, uids: Iterable[int]) -> None:
        async with self.messages_lock.write_lock():
            for uid in uids:
                try:
                    del self._messages[uid]
                except KeyError:
                    pass
            self._mod_sequences.expunge(uids)
            self._updated.set()

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        uids: List[int] = []
        async for msg in self.messages():
            if msg.recent:
                msg.recent = False
                msg_uid = msg.uid
                selected.session_flags.add_recent(msg_uid)
                uids.append(msg_uid)
        self._mod_sequences.update(uids)
        self._updated.set()

    async def cleanup(self) -> None:
        pass

    async def messages(self) -> AsyncIterable[Message]:
        async with self.messages_lock.read_lock():
            for msg in self._messages.values():
                yield msg

    async def snapshot(self) -> MailboxSnapshot:
        exists = 0
        recent = 0
        unseen = 0
        first_unseen: Optional[int] = None
        next_uid = self._max_uid + 1
        async for msg in self.messages():
            exists += 1
            if msg.recent:
                recent += 1
            if Seen not in msg.permanent_flags:
                unseen += 1
                if first_unseen is None:
                    first_unseen = exists
        return MailboxSnapshot(self.mailbox_id, self.readonly,
                               self.uid_validity, self.permanent_flags,
                               self.session_flags, exists, recent, unseen,
                               first_unseen, next_uid)


class MailboxSet(MailboxSetInterface[MailboxData]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxSetInterface`
    for the dict backend.

    """

    def __init__(self) -> None:
        super().__init__()
        self._content_cache = _ContentCache()
        self._thread_cache = _ThreadCache()
        self._inbox = MailboxData(self._content_cache, self._thread_cache)
        self._set: Dict[str, MailboxData] = OrderedDict()
        self._set_lock = subsystem.get().new_rwlock()
        self._subscribed: Dict[str, bool] = {}

    @property
    def delimiter(self) -> str:
        return '/'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        async with self._set_lock.write_lock():
            self._subscribed[name] = subscribed

    async def list_subscribed(self) -> ListTree:
        async with self._set_lock.read_lock():
            mailboxes = [child for child in self._set.keys()
                         if self._subscribed.get(child)]
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def list_mailboxes(self) -> ListTree:
        async with self._set_lock.read_lock():
            mailboxes = list(self._set.keys())
        return ListTree(self.delimiter).update('INBOX', *mailboxes)

    async def get_mailbox(self, name: str) -> MailboxData:
        if name.upper() == 'INBOX':
            return self._inbox
        async with self._set_lock.read_lock():
            return self._set[name]

    async def add_mailbox(self, name: str) -> ObjectId:
        async with self._set_lock.read_lock():
            if name in self._set:
                raise ValueError(name)
        async with self._set_lock.write_lock():
            self._set[name] = mbx = MailboxData(
                self._content_cache, self._thread_cache)
            return mbx.mailbox_id

    async def delete_mailbox(self, name: str) -> None:
        async with self._set_lock.read_lock():
            if name not in self._set:
                raise KeyError(name)
        async with self._set_lock.write_lock():
            del self._set[name]

    async def rename_mailbox(self, before: str, after: str) -> None:
        async with self._set_lock.read_lock():
            tree = ListTree(self.delimiter).update('INBOX', *self._set.keys())
            before_entry = tree.get(before)
            after_entry = tree.get(after)
            if before_entry is None:
                raise KeyError(before)
            elif after_entry is not None:
                raise ValueError(after)
        async with self._set_lock.write_lock():
            for before_name, after_name in tree.get_renames(before, after):
                if before_name == 'INBOX':
                    self._set[after_name] = self._inbox
                    self._inbox = MailboxData(
                        self._content_cache, self._thread_cache)
                else:
                    self._set[after_name] = self._set[before_name]
                    del self._set[before_name]



"""IMAP server with pluggable Python backends."""

from __future__ import annotations

import argparse
import asyncio
import logging
import logging.config
import os
from argparse import ArgumentParser, Namespace, ArgumentTypeError
from contextlib import nullcontext
from string import Template
from typing import Type, Sequence, List

from . import __version__
from .backend import backends
from .interfaces.backend import BackendInterface, ServiceInterface
from .service import services

try:
    import systemd.daemon  # type: ignore
except ImportError:
    def notify_ready() -> None:
        pass
else:
    def notify_ready() -> None:
        systemd.daemon.notify('READY=1')

try:
    from pid import PidFile  # type: ignore
except ImportError:
    def PidFile(*args, **kwargs):
        return nullcontext()

try:
    import passlib  # type: ignore
except ImportError:
    passlib = None


def main() -> None:
    parser = _PymapArgumentParser(description=__doc__)
    parser.add_argument('--debug', action='store_true',
                        help='increase printed output for debugging')
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('--set-uid', metavar='USER', type=_get_pwd,
                        help='drop privileges to user name or uid')
    parser.add_argument('--set-gid', metavar='GROUP', type=_get_grp,
                        help='drop privileges to group name or gid')
    parser.add_argument('--logging-cfg', metavar='PATH',
                        help='config file for logging')
    parser.add_argument('--no-service', dest='skip_services', action='append',
                        metavar='NAME', help='do not run the given service')
    if passlib is not None:
        parser.add_argument('--passlib-cfg', metavar='PATH',
                            help='config file for passlib hashing')
    subparsers = parser.add_subparsers(dest='backend', required=True,
                                       metavar='BACKEND')

    for backend_name, backend_type in backends:
        subparser = backend_type.add_subparser(backend_name, subparsers)
        subparser.set_defaults(run=backend_type.start)
    for _, service_type in services:
        service_type.add_arguments(parser)
    parser.set_defaults(skip_services=[], passlib_cfg=None)
    args = parser.parse_args()

    if args.logging_cfg:
        logging.config.fileConfig(args.logging_cfg)
    else:
        logging.basicConfig(level=logging.WARNING)
    if args.debug:
        logging.getLogger(__package__).setLevel(logging.DEBUG)

    backend_type = backends.registered[args.backend]
    service_types = [service for name, service in services
                     if name not in args.skip_services]

    with PidFile(force_tmpdir=True):
        try:
            return asyncio.run(run(args, backend_type, service_types),
                               debug=False)
        except KeyboardInterrupt:
            pass


async def run(args: Namespace, backend_type: Type[BackendInterface],
              service_types: Sequence[Type[ServiceInterface]]) -> None:
    backend, config = await backend_type.init(args)
    config.apply_context()

    services = [svc_type(backend, config) for svc_type in service_types]
    task = await args.run(backend, services)

    _drop_privileges(args)
    notify_ready()
    await task


def _get_pwd(setuid: str) -> int:
    from pwd import getpwnam, getpwuid
    try:
        try:
            uid = int(setuid)
        except ValueError:
            entry = getpwnam(setuid)
        else:
            entry = getpwuid(uid)
    except KeyError as exc:
        raise ArgumentTypeError(f'Invalid user: {setuid}') from exc
    else:
        return entry.pw_uid


def _get_grp(setgid: str) -> int:
    from grp import getgrnam, getgrgid
    try:
        try:
            gid = int(setgid)
        except ValueError:
            entry = getgrnam(setgid)
        else:
            entry = getgrgid(gid)
    except KeyError as exc:
        raise ArgumentTypeError(f'Invalid group: {setgid}') from exc
    else:
        return entry.gr_gid


def _drop_privileges(args: Namespace) -> None:
    if args.set_gid is not None:
        os.setgid(args.set_gid)
    if args.set_uid is not None:
        os.setuid(args.set_uid)


class _PymapArgumentParser(ArgumentParser):

    def __init__(self, **extra) -> None:
        formatter_class = argparse.ArgumentDefaultsHelpFormatter
        super().__init__(fromfile_prefix_chars='@',
                         formatter_class=formatter_class,
                         **extra)

    def convert_arg_line_to_args(self, arg_line: str) -> List[str]:
        try:
            return [Template(arg_line).substitute(os.environ)]
        except KeyError as exc:
            raise EnvironmentError(
                f'Missing environment variable: ${exc.args[0]}') from exc




from __future__ import annotations

from typing import Type

from ..interfaces.backend import BackendInterface
from ..plugin import Plugin

__all__ = ['backends']

#: Registers new backend plugins.
backends: Plugin[Type[BackendInterface]] = Plugin('pymap.backend')




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from contextlib import closing, asynccontextmanager, AsyncExitStack
from typing import Any, Type, Dict, AsyncGenerator
from typing_extensions import Final

from pymap.context import connection_exit
from pymap.exceptions import ResponseError
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.session import SessionInterface
from pymap.plugin import Plugin
from pymapadmin.grpc.admin_pb2 import Login, Result, SUCCESS, FAILURE
from pysasl import AuthenticationCredentials
from pysasl.external import ExternalResult

__all__ = ['handlers', 'BaseHandler', 'LoginHandler']

#: Registers new admin handler plugins.
handlers: Plugin[Type[BaseHandler]] = Plugin('pymap.admin.handlers')


class BaseHandler(metaclass=ABCMeta):
    """Base class for implementing admin request handlers.

    Args:
        backend: The backend in use by the system.
        public: True if the requests are received from a public-facing client.

    """

    def __init__(self, backend: BackendInterface, public: bool) -> None:
        super().__init__()
        self.config: Final = backend.config
        self.login: Final = backend.login
        self.public: Final = public

    @abstractmethod
    def __mapping__(self) -> Dict[str, Any]:
        # Remove if IServable becomes public API in grpclib
        ...

    @asynccontextmanager
    async def catch_errors(self, command: str) -> AsyncGenerator[Result, None]:
        """Context manager to catch
        :class:`~pymap.exceptions.ResponseError` exceptions and include them in
        the response.

        Args:
            command: The admin command name.

        """
        response = b'. OK %b completed.' % command.encode('utf-8')
        result = Result(code=SUCCESS, response=response)
        try:
            yield result
        except ResponseError as exc:
            result.code = FAILURE
            result.response = bytes(exc.get_response(b'.'))
            result.key = type(exc).__name__


class LoginHandler(BaseHandler, metaclass=ABCMeta):
    """Base class for implementing admin request handlers that login as a user
    to handle requests.

    Args:
        backend: The backend in use by the system.

    """

    @asynccontextmanager
    async def login_as(self, login: Login) \
            -> AsyncGenerator[SessionInterface, None]:
        """Context manager to login as a user and provide a session object.

        Args:
            login: Holds the auuthentication credentials from the request.

        Raises:
            :class:`~pymap.exceptions.InvalidAuth`

        """
        if self.public:
            creds = AuthenticationCredentials(
                login.authcid, login.secret, login.authzid or None)
        else:
            creds = ExternalResult(login.authzid or None)
        async with AsyncExitStack() as stack:
            connection_exit.set(stack)
            session = await stack.enter_async_context(self.login(creds))
            stack.enter_context(closing(session))
            yield session




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from contextlib import contextmanager, asynccontextmanager
from typing import Type, ClassVar, Optional, Tuple, Iterator, Sequence, \
    Mapping, AsyncIterator
from typing_extensions import Final, Protocol

from .bytes import BytesFormat, MaybeBytes, Writeable
from .interfaces.message import MessageInterface, LoadedMessageInterface
from .parsing.primitives import Nil, Number, ListP, LiteralString
from .parsing.specials import DateTime, FetchRequirement, FetchAttribute, \
    FetchValue
from .selected import SelectedMailbox

__all__ = ['LoadedMessageProvider', 'DynamicFetchValue',
           'DynamicLoadedFetchValue', 'MessageAttributes']

_Partial = Optional[Tuple[int, Optional[int]]]


class LoadedMessageProvider(Protocol):
    """Generic protocol that provides access to a message's loaded contents
    when they are available.

    """

    @property
    @abstractmethod
    def loaded_msg(self) -> Optional[LoadedMessageInterface]:
        """The loaded message, if available."""


class DynamicFetchValue(FetchValue, metaclass=ABCMeta):
    """Base class for fetch values that are dynamically read from a message.

    Args:
        attribute: The fetch attribute.
        message: The message object.
        selected: The selected mailbox.

    """

    __slots__ = ['message', 'selected']

    def __init__(self, attribute: FetchAttribute, *,
                 message: MessageInterface, selected: SelectedMailbox) -> None:
        super().__init__(attribute)
        self.message: Final = message
        self.selected: Final = selected

    @abstractmethod
    def get_value(self) -> MaybeBytes:
        """Computes the value of the fetch attribute for the message."""
        ...

    def __bytes__(self) -> bytes:
        return BytesFormat(b'%b %b') % (
            self.attribute.for_response, self.get_value())


class DynamicLoadedFetchValue(FetchValue, metaclass=ABCMeta):
    """Base class for fetch values that are dynamically read from a message and
    require loading its contents.

    Args:
        attribute: The fetch attribute.
        message: The message object.
        selected: The selected mailbox.
        get_loaded: The provider of the loaded message contents, if available.

    """

    __slots__ = ['message', '_get_loaded']

    def __init__(self, attribute: FetchAttribute, *,
                 message: MessageInterface,
                 get_loaded: LoadedMessageProvider) -> None:
        super().__init__(attribute)
        self.message: Final = message
        self._get_loaded = get_loaded

    @abstractmethod
    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        """Computes the value of the fetch attribute for the message.

        Args:
            loaded_msg: The loaded message object.

        """
        ...

    def __bytes__(self) -> bytes:
        loaded_msg = self._get_loaded.loaded_msg
        if loaded_msg is None:
            value: MaybeBytes = MessageAttributes.placeholder
        else:
            value = self.get_value(loaded_msg)
        return BytesFormat(b'%b %b') % (
            self.attribute.for_response, value)

    @classmethod
    def _get_data(cls, section: FetchAttribute.Section, partial: _Partial,
                  loaded_msg: LoadedMessageInterface, *,
                  binary: bool = False) -> Writeable:
        specifier = section.specifier
        parts = section.parts
        headers = section.headers
        if specifier is None:
            data = loaded_msg.get_body(parts, binary)
        elif specifier == b'MIME' and parts is not None:
            data = loaded_msg.get_headers(parts)
        elif specifier == b'TEXT':
            data = loaded_msg.get_message_text(parts)
        elif specifier == b'HEADER':
            data = loaded_msg.get_message_headers(parts)
        elif specifier == b'HEADER.FIELDS':
            data = loaded_msg.get_message_headers(parts, headers)
        elif specifier == b'HEADER.FIELDS.NOT':
            data = loaded_msg.get_message_headers(parts, headers, True)
        else:
            raise RuntimeError()  # Should not happen.
        return cls._get_partial(data, partial)

    @classmethod
    def _get_partial(cls, data: Writeable, partial: _Partial) -> Writeable:
        if partial is None:
            return data
        full = bytes(data)
        start, end = partial
        if end is None:
            end = len(full)
        return Writeable.wrap(full[start:end])


class _UidFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        return Number(self.message.uid)


class _FlagsFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        session_flags = self.selected.session_flags
        flag_set = self.message.get_flags(session_flags)
        return ListP(flag_set, sort=True)


class _InternalDateFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        return DateTime(self.message.internal_date)


class _EmailIdFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        return self.message.email_id.parens


class _ThreadIdFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        try:
            return self.message.thread_id.parens
        except ValueError:
            return Nil()


class _LoadedMessageProvider(LoadedMessageProvider):

    __slots__ = ['loaded_msg']

    def __init__(self) -> None:
        super().__init__()
        self.loaded_msg: Optional[LoadedMessageInterface] = None

    @contextmanager
    def apply(self, loaded_msg: LoadedMessageInterface) -> Iterator[None]:
        self.loaded_msg = loaded_msg
        try:
            yield
        finally:
            self.loaded_msg = None


class _EnvelopeFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        return loaded_msg.get_envelope_structure()


class _BodyStructureFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        return loaded_msg.get_body_structure().extended


class _BodyFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        attr = self.attribute
        if attr.section is None:
            return loaded_msg.get_body_structure()
        else:
            data = self._get_data(attr.section, attr.partial, loaded_msg)
            return LiteralString(data)


class _BodyPeekFetchValue(_BodyFetchValue):
    pass


class _RFC822FetchValue(_BodyFetchValue):
    pass


class _RFC822HeaderFetchValue(_BodyFetchValue):
    pass


class _RFC822TextFetchValue(_BodyFetchValue):
    pass


class _RFC822SizeFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        return Number(loaded_msg.get_size())


class _BinaryFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        attr = self.attribute
        if attr.section is None:
            raise RuntimeError()  # should not happen.
        data = self._get_data(attr.section, attr.partial, loaded_msg,
                              binary=True)
        return LiteralString(data, True)


class _BinaryPeekFetchValue(_BinaryFetchValue):
    pass


class _BinarySizeFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        attr = self.attribute
        if attr.section is None:
            raise RuntimeError()  # should not happen.
        data = self._get_data(attr.section, attr.partial, loaded_msg,
                              binary=True)
        return Number(len(data))


class MessageAttributes(Sequence[FetchValue]):
    """Defines the logic for how fetch attributes are resolved on a message to
    produce a fetch value.

    Args:
        message: The message object.
        selected: The selected mailbox.

    """

    #: Placeholder value for fetch values requiring loaded message contents.
    placeholder: ClassVar[bytes] = b'...'

    _simple_attrs: Mapping[bytes, Type[DynamicFetchValue]] = {
        b'UID': _UidFetchValue,
        b'FLAGS': _FlagsFetchValue,
        b'INTERNALDATE': _InternalDateFetchValue,
        b'EMAILID': _EmailIdFetchValue,
        b'THREADID': _ThreadIdFetchValue}

    _loaded_attrs: Mapping[bytes, Type[DynamicLoadedFetchValue]] = {
        b'ENVELOPE': _EnvelopeFetchValue,
        b'BODYSTRUCTURE': _BodyStructureFetchValue,
        b'BODY': _BodyFetchValue,
        b'BODY.PEEK': _BodyPeekFetchValue,
        b'RFC822': _RFC822FetchValue,
        b'RFC822.HEADER': _RFC822HeaderFetchValue,
        b'RFC822.TEXT': _RFC822TextFetchValue,
        b'RFC822.SIZE': _RFC822SizeFetchValue,
        b'BINARY': _BinaryFetchValue,
        b'BINARY.PEEK': _BinaryPeekFetchValue,
        b'BINARY.SIZE': _BinarySizeFetchValue}

    __slots__ = ['message', 'selected', 'attributes', 'requirement',
                 '_get_loaded', '_values']

    def __init__(self, message: MessageInterface,
                 selected: SelectedMailbox,
                 attributes: Sequence[FetchAttribute]) -> None:
        super().__init__()
        self.message: Final = message
        self.selected: Final = selected
        self.attributes: Final = attributes
        self.requirement: Final = FetchRequirement.reduce(
            attr.requirement for attr in attributes)
        self._get_loaded = _LoadedMessageProvider()
        self._values: Optional[Sequence[FetchValue]] = None

    @asynccontextmanager
    async def load_hook(self) -> AsyncIterator[None]:
        """An async context manager that loads the message content on entry,
        and releases it on exit. Fetch values that require loaded message
        content will write :attr:`.placeholder` if written outside of this
        context manager, for console or log output.

        """
        loaded_msg = await self.message.load_content(self.requirement)
        with self._get_loaded.apply(loaded_msg):
            yield

    def __iter__(self) -> Iterator[FetchValue]:
        return iter(self._get_values())

    def __getitem__(self, index):
        values = self._get_values()
        return values[index]

    def __len__(self) -> int:
        return len(self.attributes)

    def _get_values(self) -> Sequence[FetchValue]:
        if self._values is None:
            self._values = [self._get(attr) for attr in self.attributes]
        return self._values

    def _get(self, attr: FetchAttribute) -> FetchValue:
        attr_name = attr.value
        simple = self._simple_attrs.get(attr_name)
        if simple is not None:
            return simple(attr, message=self.message, selected=self.selected)
        loaded = self._loaded_attrs.get(attr_name)
        if loaded is not None:
            return loaded(attr, message=self.message,
                          get_loaded=self._get_loaded)
        raise KeyError(attr_name)




from __future__ import annotations

from grpclib.server import Stream
from pymap import __version__ as server_version
from pymapadmin.grpc.admin_grpc import SystemBase
from pymapadmin.grpc.admin_pb2 import PingRequest, PingResponse

from . import BaseHandler

__all__ = ['SystemHandlers']

_PingStream = Stream[PingRequest, PingResponse]


class SystemHandlers(SystemBase, BaseHandler):
    """The GRPC handlers, executed when an admin request is received. Each
    handler should receive a request, take action, and send the response.

    See Also:
        :class:`grpclib.server.Server`

    """

    async def Ping(self, stream: _PingStream) -> None:
        """Respond to a ping request. For example::

            $ pymap-admin ping

        See ``pymap-admin ping --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        request = await stream.recv_message()
        assert request is not None
        resp = PingResponse(server_version=server_version,
                            public_client=self.public)
        await stream.send_message(resp)




from __future__ import annotations

from abc import abstractmethod
from typing import Collection, AsyncIterable
from typing_extensions import Protocol

from ..user import UserMetadata

__all__ = ['UsersInterface']


class UsersInterface(Protocol):
    """Defines admin functions that backends can implement to manage users."""

    @abstractmethod
    def list_users(self, *, match: str = None) \
            -> AsyncIterable[Collection[str]]:
        """Iterate all matching users in batches. The format *match* argument
        depends on the backend implementation.

        Args:
            match: A filter string for matched users.

        """
        ...

    @abstractmethod
    async def get_user(self, user: str) -> UserMetadata:
        """Return the password and other metadata for a username.

        Args:
            user: The user login string.

        Raises:
            :class:`~pymap.exceptions.UserNotFound`

        """
        ...

    @abstractmethod
    async def set_user(self, user: str, data: UserMetadata) -> None:
        """Assign a password and other metadata to a username, creating it if
        it does not exist.

        Args:
            user: The user login string.
            data: The user metadata, including password.

        """
        ...

    @abstractmethod
    async def delete_user(self, user: str) -> None:
        """Delete a user and all mail data associated with it.

        Args:
            user: The user login string.

        Raises:
            :class:`~pymap.exceptions.UserNotFound`

        """
        ...




from __future__ import annotations

from typing import Tuple

from . import CommandNonAuth, CommandNoArgs
from .. import Params, Space, EndLine
from ..primitives import Atom
from ..specials import AString

__all__ = ['AuthenticateCommand', 'LoginCommand', 'StartTLSCommand']


class AuthenticateCommand(CommandNonAuth):
    """The ``AUTHENTICATE`` command authenticates an IMAP session using a SASL
    mechanism.

    Args:
        tag: The command tag.
        mech_name: The SASL mechanism name.

    """

    command = b'AUTHENTICATE'

    def __init__(self, tag: bytes, mech_name: bytes) -> None:
        super().__init__(tag)
        self.mech_name = mech_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[AuthenticateCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        atom, after = Atom.parse(buf, params)
        _, after = EndLine.parse(after, params)
        return cls(params.tag, atom.value.upper()), after


class LoginCommand(CommandNonAuth):
    """The ``LOGIN`` command authenticates an IMAP session using a basic user
    ID and password credentials in clear-text.

    Args:
        tag: The command tag.
        userid: The user ID bytestring.
        password: The password bytestring.

    """

    command = b'LOGIN'

    def __init__(self, tag: bytes, userid: bytes, password: bytes) -> None:
        super().__init__(tag)
        self.userid = userid
        self.password = password

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[LoginCommand, memoryview]:
        _, buf = Space.parse(buf, params)
        userid, buf = AString.parse(buf, params)
        _, buf = Space.parse(buf, params)
        password, buf = AString.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag, userid.value, password.value), buf


class StartTLSCommand(CommandNoArgs, CommandNonAuth):
    """The ``STARTTLS`` command attempts to encrypt a non-encrypted IMAP
    session using opportunistic TLS. The client/server handshake should take
    place immediately after the server issues a
    :class:`~pymap.parsing.response.ResponseOk`.

    """

    command = b'STARTTLS'



# Copyright (c) 2020 Ian C. Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from setuptools import setup, find_packages  # type: ignore

with open('README.md') as f:
    readme = f.read()

with open('LICENSE.md') as f:
    license = f.read()

setup(name='pymap',
      version='0.20.1',
      author='Ian Good',
      author_email='ian@icgood.net',
      description='Lightweight, asynchronous IMAP serving in Python.',
      long_description=readme + license,
      long_description_content_type='text/markdown',
      license='MIT',
      url='https://github.com/icgood/pymap/',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Topic :: Communications :: Email :: Post-Office',
          'Topic :: Communications :: Email :: Post-Office :: IMAP',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.8'],
      python_requires='~=3.8',
      include_package_data=True,
      packages=find_packages(),
      install_requires=[
          'pysasl >= 0.6.1',
          'proxy-protocol >= 0.5.0',
          'typing-extensions'],
      extras_require={
          'redis': ['aioredis >= 1.3.0', 'msgpack >= 1.0'],
          'admin': ['pymap-admin == 0.4.2'],
          'sieve': ['sievelib'],
          'systemd': ['systemd-python'],
          'optional': ['hiredis', 'passlib', 'pid']},
      entry_points={
          'console_scripts': [
              'pymap = pymap.main:main'],
          'pymap.backend': [
              'dict = pymap.backend.dict:DictBackend',
              'maildir = pymap.backend.maildir:MaildirBackend',
              'redis = pymap.backend.redis:RedisBackend [redis]'],
          'pymap.service': [
              'imap = pymap.imap:IMAPService',
              'admin = pymap.admin:AdminService [admin]',
              'managesieve = pymap.sieve.manage:ManageSieveService [sieve]'],
          'pymap.filter': [
              'sieve = pymap.sieve:SieveCompiler [sieve]'],
          'pymap.admin.handlers': [
              'server = pymap.admin.handlers.system:SystemHandlers',
              'mailbox = pymap.admin.handlers.mailbox:MailboxHandlers',
              'user = pymap.admin.handlers.user:UserHandlers']})




from __future__ import annotations

from typing_extensions import Final

from pymap.config import IMAPConfig
from pymap.interfaces.filter import FilterSetInterface

from .command import Command, HaveSpaceCommand, PutScriptCommand, \
    ListScriptsCommand, SetActiveCommand, GetScriptCommand, \
    DeleteScriptCommand, RenameScriptCommand, CheckScriptCommand
from .response import Condition, Response, GetScriptResponse, \
    ListScriptsResponse
from .. import SieveParseError

__all__ = ['FilterState']


class FilterState:

    def __init__(self, filter_set: FilterSetInterface[bytes],
                 owner: bytes, config: IMAPConfig) -> None:
        super().__init__()
        self.filter_set: Final = filter_set
        self.owner: Final = owner
        self.config: Final = config

    async def _do_have_space(self, cmd: HaveSpaceCommand) -> Response:
        max_len = self.config.max_filter_len
        if max_len is None or cmd.size <= max_len:
            return Response(Condition.OK)
        else:
            return Response(Condition.NO, code=b'QUOTA/MAXSIZE')

    async def _do_put_script(self, cmd: PutScriptCommand) -> Response:
        max_len = self.config.max_filter_len
        if max_len is None or len(cmd.script_data) <= max_len:
            await self.filter_set.put(cmd.script_name, cmd.script_data)
            return Response(Condition.OK)
        else:
            return Response(Condition.NO, code=b'QUOTA/MAXSIZE')

    async def _do_list_scripts(self) -> Response:
        active, names = await self.filter_set.get_all()
        return ListScriptsResponse(active, names)

    async def _do_set_active(self, cmd: SetActiveCommand) -> Response:
        if cmd.script_name is None:
            await self.filter_set.clear_active()
        else:
            try:
                await self.filter_set.set_active(cmd.script_name)
            except KeyError:
                return Response(Condition.NO, code=b'NONEXISTENT')
        return Response(Condition.OK)

    async def _do_get_script(self, cmd: GetScriptCommand) -> Response:
        try:
            script_data = await self.filter_set.get(cmd.script_name)
        except KeyError:
            return Response(Condition.NO, code=b'NONEXISTENT')
        return GetScriptResponse(script_data)

    async def _do_delete_script(self, cmd: DeleteScriptCommand) -> Response:
        try:
            await self.filter_set.delete(cmd.script_name)
        except KeyError:
            return Response(Condition.NO, code=b'NONEXISTENT')
        except ValueError:
            return Response(Condition.NO, code=b'ACTIVE')
        else:
            return Response(Condition.OK)

    async def _do_rename_script(self, cmd: RenameScriptCommand) -> Response:
        try:
            await self.filter_set.rename(cmd.old_script_name,
                                         cmd.new_script_name)
        except KeyError as exc:
            if exc.args == (cmd.old_script_name, ):
                return Response(Condition.NO, code=b'NONEXISTENT')
            elif exc.args == (cmd.new_script_name, ):
                return Response(Condition.NO, code=b'ALREADYEXISTS')
            else:
                return Response(Condition.NO)
        else:
            return Response(Condition.OK)

    async def _do_check_script(self, cmd: CheckScriptCommand) -> Response:
        try:
            await self.filter_set.compiler.compile(cmd.script_data)
        except SieveParseError as exc:
            return Response(Condition.NO, text=str(exc))
        else:
            return Response(Condition.OK)

    async def run(self, cmd: Command) -> Response:
        try:
            if isinstance(cmd, HaveSpaceCommand):
                return await self._do_have_space(cmd)
            elif isinstance(cmd, PutScriptCommand):
                return await self._do_put_script(cmd)
            elif isinstance(cmd, ListScriptsCommand):
                return await self._do_list_scripts()
            elif isinstance(cmd, SetActiveCommand):
                return await self._do_set_active(cmd)
            elif isinstance(cmd, GetScriptCommand):
                return await self._do_get_script(cmd)
            elif isinstance(cmd, DeleteScriptCommand):
                return await self._do_delete_script(cmd)
            elif isinstance(cmd, RenameScriptCommand):
                return await self._do_rename_script(cmd)
            elif isinstance(cmd, CheckScriptCommand):
                return await self._do_check_script(cmd)
            else:
                return Response(Condition.NO, text='Bad command.')
        except NotImplementedError:
            return Response(Condition.NO, text='Action not supported.')
